import com.revusky.oreo.*;
import com.revusky.niggle.servlet.*;
import javax.servlet.http.*;
import java.io.IOException;
import java.util.List;


/**
 * The code that provides the dispatching
 * logic for our first pass at a rolodex
 */
public class MiniRoloServletInteraction extends ServletInteraction {

    /**
     * This constructor must have exactly this signature, since
     * Niggle invokes it by reflective lookup.
     */
    public MiniRoloServletInteraction(HttpServletRequest req, 
                                      HttpServletResponse resp, 
                                      NiggleConfig config) 
    throws IOException {
        super(req, resp, config);
    }

    /**
     * Displays our default view, which is simply
     * a view of all the entries. (unsorted)
     */
    public void execDefault() throws IOException
    {
        MutableDataSource data = (MutableDataSource) getDataSource("rolodex_data");
        List entries = data.select(null); 
        page = getPage("entries.nhtml");
        page.expose("entries", entries);
    }

    /**
     * deletes the record indicated by the unique_id
     * parameter and displays the acknowledgement page.
     */
    public void execDelete() throws IOException {
        MutableDataSource data = (MutableDataSource) getDataSource("rolodex_data");
        Integer recordID = getEntryID();
        Record entry = data.get("rolodex_entry", recordID);
        if (entry == null)
            throw new MissingRecordException();
        data.delete(entry);
        page = getPage("ack.nhtml");
        page.expose("operation", "deleted");
        page.expose("entry", entry);
    }

    /**
     * displays a page that allows
     * the user to edit the fields of a rolodex entry.
     */
    public void execEdit() throws IOException {
        Record entry = null;
        Integer recordID = getEntryID();
        if (recordID != null)  {
            entry = getDataSource("rolodex_data").get("rolodex_entry", recordID);
            if (entry == null)
                throw new MissingRecordException("No entry with unique id: " + recordID);
        }
        else {
            entry = getNewRecord("rolodex_entry");
        }
        page = getPage("edit.nhtml");
        page.expose("entry", entry);
    }

    /**
     * Processes the form embedded in the edit page template
     * and displays the acknowledgement page.
     */
    public void execProcess() throws IOException {
        String operation = null;
        Record newEntry = null;
        MutableDataSource mds = (MutableDataSource) getDataSource("rolodex_data");
        Integer recordID = getEntryID();
        if (recordID == null) {
            newEntry = getNewRecord("rolodex_entry");
            fillInFields(newEntry);
            mds.insert(newEntry);
            operation = "added";
        }
        else {
            Record oldEntry = mds.get("rolodex_entry", recordID);
            if (oldEntry == null) 
                throw new MissingRecordException();
            newEntry = oldEntry.getMutableCopy();
            fillInFields(newEntry);
            mds.update(oldEntry, newEntry);
            operation = "modified";
        }
        page = getPage("ack.nhtml");
        page.expose("operation", operation);
        page.expose("entry", newEntry);
    }

    private Integer getEntryID() {
        String s = getParameter("unique_id");
        int id = 0;
        boolean hasNumber = false;
        for (int i=0; i<s.length(); i++) {
            char c = s.charAt(i);
            if (Character.isDigit(c)) {
                id = id * 10 + (c - '0');
                hasNumber = true;
            }
        }
        return hasNumber ? new Integer(id) : null;
    }

    /**
     * a convenient hook for handling error conditions.
     */
    protected void recover(IOException e) throws IOException {
        try {
            throw e;
        }
        catch (DataException de) {
            Record entry = de.getRecord();
            if (entry != null) {
                page = getPage("edit.nhtml");
                page.expose("entry", entry);
            }
            else {
                page = getPage("ack.nhtml");
            }
            page.expose("error", de.getLocalizedMessage());
        }
    }

    /**
     * A convenient place to define certain variables
     * that will always be exposed, independently
     * of the action that was executed.
     */
    protected void exposeDefaultVariables() throws IOException {
        super.exposeDefaultVariables();
        exposeActionURL("default");
        exposeActionURL("edit");
        exposeActionURL("delete");
        exposeActionURL("process");
    }
}

