/*
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned 
by Reminders On-Line, Inc. It was released under the terms 
of the BSD-style open source license below. The terms of that 
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL) 
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.niggle.templates.freemarkerimpl;

import java.io.*;
import java.util.*;
import javax.servlet.ServletContext;
import freemarker.template.*;
import com.revusky.oreo.util.Util;
import com.revusky.niggle.templates.PresentationWrapper;

/**
 * A version of the Freemarker TemplateCache that fishes out a
 * different template based on locale information.
 * (should be rolled into Freemarker actually)
 * @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
 */

public class LocalizedFileTemplateCache extends FileTemplateCache
implements LocalizedTemplateCache {

    public LocalizedFileTemplateCache() {
    }

    public LocalizedFileTemplateCache(String path, ClassLoader cl) {
        if ("".equals(path)) 
            path = null;
        setPath(path);
        setClassLoader(cl);
    }


    /**
     * fish out the Template object based on the given base filename
     * and locale.
     */
    public Template getTemplate(String filename, Locale locale) {
        String[] fnames = PresentationWrapper.getFilenames(filename, locale);
        Template result = lookInContext(fnames, locale);
        if (result != null) {
            result.setTemplateCache(this);
        }
        return result;
    }

    public Template getTemplate(String filename) {
        Template result = super.getTemplate(filename);
        if (result == null) {
            result = getTemplate(filename, Locale.getDefault());
        }
        return result;
    }


    /**
     * Run through the list of filenames, which
     * are assumed to be listed in order of more
     * specifically qualified down to default,
     * and try to get the resource relative to
     * the CLASSPATH.
     * @return null on failure.
     */
    private Template lookInContext(String[] fnames, Locale locale) {
        for (int i = 0; i < fnames.length; i++) {
            if (fnames[i] == null)
                continue;
            if (explicitDirectory != null) {
                Template tm = getTemplate(fnames[i]);
                if (tm != null)
                    return tm;
            }
            String filename = fnames[i];
            if (path != null) {
                filename = path + "/" + filename;
            }
            Date d = (Date) lastLoadedCache.get(filename);
            if (d != null) {
                long l = System.currentTimeMillis() - d.getTime();
                if (l > updateFrequency) {
                    ourTemplateCache.remove(filename);
                }
            }
            if (ourTemplateCache.containsKey(filename)) {
                Template tm = (Template) ourTemplateCache.get(filename);
                if (tm == null)
                    continue;
                return tm;
            }
            InputStream is = Util.getInputStream(filename, cl);
            if (is == null) {
                // Okay, stick a null in so we won't waste
                // our time next time!.
                ourTemplateCache.put(filename, null);
                lastLoadedCache.put(filename, new Date());
            }
            else try {
                    if (i ==2)
                        locale = new Locale(locale.getLanguage(), "");
                    else if (i == 3)
                        locale = Locale.getDefault();
                    Template tm = new LocalizedTemplate(is, locale); 
                    ourTemplateCache.put(filename, tm);
                    lastLoadedCache.put(filename, new Date());
                    return tm;
                }
                catch (Exception e) {
                    ourTemplateCache.put(filename, null);
                    lastLoadedCache.put(filename, new Date());
                }
        }
        return null; // Couldn't find anything!
    }

    public void setPath(String path) {
        this.path = path;
        if (path != null) {
            File dir = new File(path);
            char firstChar = 'a';
            if (path.length() >0)
                firstChar = path.charAt(0);
            boolean absolute = dir.isAbsolute() || firstChar=='/' || firstChar=='\\';
            if (absolute && dir.canRead()) {
                this.explicitDirectory = dir;
                super.setDirectory(dir);
                startAutoUpdate();
            }
        }
    }

    void setClassLoader(ClassLoader cl) {
        this.cl = cl;
    }


    private final Map ourTemplateCache = Collections.synchronizedMap(new HashMap());
    private final Map lastLoadedCache = Collections.synchronizedMap(new HashMap());
    long updateFrequency = 5000; // later make configurable
    private String path = null;
    private File explicitDirectory;
    private ClassLoader cl;
}
