/*
Bridge Code to use WebMacro template engine with Niggle
Copyright (c) 2001, Jonathan Revusky
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice, 
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation 
and/or other materials provided with the distribution.

The name of Jonathan Revusky may not be used to endorse or promote 
products derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE. 
*/

package com.revusky.niggle.templates.webmacroimpl;

import java.io.*;
import java.util.*;
import java.net.URL;
import javax.servlet.*;
import com.revusky.niggle.servlet.*;
import com.revusky.niggle.templates.*;
import com.revusky.oreo.util.*;
import org.webmacro.*;


/**
 * A concrete PageFactory type that vends Page objects
 * that wrap WebMacro template objects
 * @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
 */

public class WebMacroPageFactory implements PageFactory {

    private WM wm;
    protected ClassLoader classLoader;
    String templatePath;

    public Page getPage(String name, ServletInteraction interaction) throws IOException {
        Locale locale = interaction.getLocale();
        Template tmp = null;
        String[] filenames = PresentationWrapper.getFilenames(name, locale);
        String encoding = PresentationWrapper.getCharset(locale);
        for (int i = 0; i<filenames.length; i++) {
            if (i==3) {
                locale = Locale.getDefault();
                encoding = PresentationWrapper.getCharset(locale);
            }
            tmp = getTemplate(filenames[i], encoding);
            if (tmp != null) {
                break;
            }
        }
        Context ct = wm.getWebContext(interaction.request, interaction.response);
        return tmp != null ?
            new WebMacroPage(tmp, ct, locale, classLoader) :
            null;
        }

    Template getTemplate(String filename, String encoding) {
        WebMacroTemplate result = (WebMacroTemplate) templateCache.get(filename);
        if (result != null) {
            result.reparse(updateFrequency);
        }
        else  {
            result = (WebMacroTemplate) findTemplate(filename, encoding);
        }
        return result;
    }

    WebMacroTemplate findTemplate(String filename, String encoding)  {
        if (!Util.hasContent(filename)) {
            return null;
        }
        URL resource = null;
        if (Util.hasContent(templatePath)) {
            StringTokenizer st = new StringTokenizer(templatePath, File.pathSeparator);
            while (st.hasMoreTokens()) {
                String path = st.nextToken();
                char lastChar = path.charAt(path.length() -1);
                if (lastChar == '/' || lastChar == File.separatorChar) {
                    path = path.substring(0, path.length() -1);
                }
                resource = getResource(path + '/' + filename);
                if (resource != null) {
                    break;
                }
            }
        }
        if (resource == null) {
            resource = getResource(filename);
        }
        WebMacroTemplate result = null;
        if (resource != null) {
            result = new WebMacroTemplate(filename, wm.getBroker(), resource, encoding);
        }
        templateCache.put(filename, result);
        return result;
    }

    URL getResource(String location) {
        URL result = null;
        File f = new File(location);
        if (f.exists()) {
            try {
                return  f.toURL();
            }
            catch (Exception e) {
                Assert.rethrow(e);
            }
        }
        result = classLoader.getResource(location);
        if (result == null) {
            ClassLoader cl = this.getClass().getClassLoader();
            if (cl != null && cl != classLoader) {
                result = cl.getResource(location);
            }
        }
        if (result == null)
            result = ClassLoader.getSystemResource(location);
        return result;
    }

    public void setConfig(NiggleConfig config) throws ServletException {
        this.classLoader = config.getServletClassLoader();
        Servlet servlet = config.getServlet();
        if (servlet != null)  try {
                wm = new WM(servlet);
            }
            catch (InitException e) {
                throw new ServletException(e);
            }
        templatePath = config.getInitParameter("RESOURCE_BASE");
        if (!Util.hasContent(templatePath)) {
            templatePath = getWebMacroSetting("TemplatePath");
        }
    }

    String getWebMacroSetting(String key) {
        return wm.getBroker().getSetting(key);
    }

    private Map templateCache = Collections.synchronizedMap(new HashMap());
    private long updateFrequency = 5000L;
}
