/*
OREO Object-relational layer
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned
by Reminders On-Line, Inc. It was released under the terms
of the BSD-style open source license below. The terms of that
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.oreo;

import com.revusky.oreo.util.*;
import java.io.*;
import java.util.*;

/**
 *  A convenient base class for MutableDataSource implementations.
 *  Implements some methods that are likely to be used by most.
 *
 *  @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
 */

abstract public class AbstractMutableDataSource extends AbstractDataSource
implements MutableDataSource {

    /**
     * Our constructor must throw the exception
     * since the superclass's constructor does.
     */
    public AbstractMutableDataSource() throws IOException {
    }

    public void addDataListener(DataListener dl) throws IOException {
        if ( listeners == null ) {
            listeners = new ArrayList();
        }
        listeners.remove(dl);
        listeners.add(dl);
    }

    public void removeDataListener(DataListener dl) throws IOException {
        if ( listeners != null ) {
            listeners.remove(dl);
            if ( listeners.size() == 0 ) {
                listeners = null;
            }
        }
    }

    /**
     * Fire a data event to the listeners.
     */
    protected void fireDataEvent(DataEvent event) {
        if ( listeners == null )
            return;
        List l = (List) listeners.clone();
        for (Iterator it = l.iterator(); it.hasNext();) {
            try {
                ((DataListener) it.next()).handleEvent(event);
            }
            catch ( IOException ioe ) {
                Logger.logException(ioe, Logger.LOG_WARN);
            }
        }
    }

    /**
     * Called to mark the record as immutable.
     * A subclass might override this routine, particularly
     * one that uses a different implementation of the
     * Record interface.
     * @param rec the Record to freeze
     * @param validate You can make this true if you don't want validation to take place.
     * Be careful with this. You might use this if you are absolutely sure that there
     * is no way that the data you are slurping in could be invalid. Thus you avoid
     * superfluous checks.
     */
    protected void freeze(Record rec, boolean validate) throws IOException {
        if (!validate) {
            try {
                ((DefaultRecord) rec).immutable = true;
            }
            catch (ClassCastException cce) {
                rec.freeze();
            }
        }
        else
            rec.freeze();
    }

    /**
     * This is equivalent to calling freeze(rec, true);
     */
    protected final void freeze(Record rec) throws IOException {
        freeze(rec, true);
    }

    /**
     * Mark the record as "stale". It has either been
     * deleted or superseded by a newer version.
     */
    protected void markStale(Record rec) {
        ((DefaultRecord) rec).markStale();
    }

    /**
     * @return the "normalized" key given the record type.
     */

    static protected Object normalizeKey(String type, String key) throws IOException {
        String result = key;
        if (Util.hasContent(type)) {
            DataRegistry reg = DataRegistryImpl.getDefaultRegistry();
            RecordDescriptor descriptor = reg.getMetadata(type);
            FieldDescriptor primaryKeyField = descriptor.getPrimaryKeyField();
            result = primaryKeyField.normalize(result);
            if (primaryKeyField.getJavaType() != result.getClass()) {
                return primaryKeyField.getObjectFromString(null, result);
            }
        }
        return result;
    }


    public final List getRecords(Object key) throws IOException {
        ArrayList result = new ArrayList(1);
        Record rec = get(key);
        if (rec != null)
            result.add(rec);
        return result;
    }

    /**
     * Listeners to inform of changes to the data in this container.
     */
    ArrayList listeners;
}
