/*
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned
by Reminders On-Line, Inc. It was released under the terms
of the BSD-style open source license below. The terms of that
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.oreo;

import java.io.*;
import java.util.*;
import java.text.*;
import javax.servlet.*;
import com.revusky.oreo.util.*;


/**
 * a holder for various useful static routines dealing
 * with Oreo data
 * @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
 */

abstract public class DataUtil {

    /**
     * convenience routine for transitioning records.
     * Copies all the fields in source to dest,
     * where dest has a field of the same name.
     */
    static public void copyFieldsByName(Record source, Record dest) {
        Assert.assert(!dest.isImmutable());
        RecordDescriptor sourceFields = source.getMetadata();
        RecordDescriptor destFields = dest.getMetadata();
        for (Enumeration e = sourceFields.getFieldDescriptors(); e.hasMoreElements();) {
            FieldDescriptor sourceField = (FieldDescriptor) e.nextElement();
            FieldDescriptor destField = (FieldDescriptor) destFields.getField(sourceField.getName());
            if (destField != null) {
                try {
                    Object val = source.get(sourceField);
                    dest.set(destField, val);
                }
                catch (InvalidFieldException ife) {
                    // if we can't do it, we can't do it.
                }
            }
        }
    }

    /**
     * Read list of key-value pairs representing one record in the object stream.
     * List is terminated by a blank line or EOF.
     * Ignores #comment lines.
     * @param rec the Record involved. We need this to get at the RecordDescriptor
     * metadata object.
     * @param in     Stream from which to read the record
     * @return a table of key-value pairs from the input stream, i.e. a HashTable
     * @exception IOException
     */
    static Map getFieldValueTable(Record rec, DataInput in) throws IOException {
        final HashMap result = new HashMap();
        String line = null;
        do {
            line = in.readLine();
            if ( line == null )
                break;
            line = line.trim();
        }
        while ( line.length() == 0  || line.charAt(0) == '#' );

        // okay, now we should have got the first line.

        while ( line != null && line.length() >0 ) {
            if ( line.charAt(0) != '#' ) {
                if (Util.hasNonAscii(line))
                    line = Util.fromByteEncodedString(line);
                int index = line.indexOf('=');
                if ( index < 0 )
                    throw new MangledDataException(rec, "invalid format, no equals sign delimiter");
                String key = line.substring(0, index).toLowerCase().intern();
                String value = line.substring(++index);
                result.put(key, value);
            }
            line = in.readLine();
            if ( line != null ) {
                line = line.trim();
            }
        }
        return result;
    }

    /**
     * convenience routine to convert a string to a Record.
     * Is essentially the reverse operation of: DefaultRecord.toString()
     */
    public static Record fromString(String s) throws IOException, ClassNotFoundException {
        ByteArrayInputStream bais = new ByteArrayInputStream(s.getBytes());
        ObjectInput oi = new DefaultRecordInputStream(bais);
        return(Record) oi.readObject();
    }
}
