/*
OREO Object-relational layer
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned
by Reminders On-Line, Inc. It was released under the terms
of the BSD-style open source license below. The terms of that
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.oreo;

import com.revusky.oreo.util.*;
import java.io.*;
import java.util.*;

/**
 * A concrete implementation of the Record interface
 * In this implementation, the records can be written out
 * as a human-readable string. This is the scheme used
 * for serialization. DefaultRecord implements the
 * java.io.Externalizable interface in order to avoid
 * the general-purpose serialization scheme. The goal was that the
 * resulting storage format should be human-readable and
 * modifiable in a text editor in a pinch.
 * @author <a href="mailto:jon@revusky.com>Jonathan Revusky</a>
 */

public class DefaultRecord implements Record, Externalizable, Cloneable {

    /**
     * The values in this record.
     */
    private Object[] values;

    /**
     * The metadata object that describes the
     * contents of this record.
     */
    private RecordDescriptor descriptor;

    private String type;

    /**
     *  Flags to indicate whether a record has been
     * deleted or superseded, and whether the record is in an immutable state.
     */
    private boolean stale;

    boolean immutable;

    /**
     *  Magic number used by the serialization mechanism
     */
    static final long serialVersionUID = -4056191582361765908L;

    /**
     * Once a record is immutable,
     * attempts to change data will result in
     * an ImmutableDataException being thrown.
     * @return boolean value representing whether the
     * record is immutable.
     */
    public boolean isImmutable() {
        return immutable;
    }

    /**
     *  Check if the record's data is valid.
     *
     *  @exception IOException if any of the fields are not valid.
     *  The appropriate error message is encoded in the exception.
     */
    public final void checkValidity() throws IOException {
        if ( isImmutable() )
            return;
        preCheck();
        checkFields();
        postCheck();
    }

    /**
     * Called to indicate that the data in this record
     * is no longer relevant. Really, in our overall scheme,
     * only a MutableDataSource should call this, typically in
     * its delete() method.
     */
    void markStale() {
        stale = true;
    }

    public boolean isStale() {
        return stale;
    }

    public Object getPrimaryKey() {
// This is a possible speedup assuming the primary key is an immutable type. REVISIT
        // return values[0];
        return get(getMetadata().getPrimaryKeyField());
    }

    public void setPrimaryKey(Object o) {
        if (immutable)
            throw new ImmutableDataException();
        Object pKey = getPrimaryKey();
        // mini-kludge so that the empty string
        // is treated symmetrically with null.
        if (pKey != null && pKey.equals(o))
            return;
        if (pKey != null && o != null) {
            throw new ImmutableDataException("The primary key cannot be " +
                                             "modified once set." +
                                             " Old key: " + pKey +
                                             ";" +
                                             " New key: " + o + ".");
        }
        set(getMetadata().getPrimaryKeyField(), o);
    }

    /**
     * check that all fields in this Record have valid values.
     * Each field type has a custom checkValidValue method.
     */
    final void checkFields() throws IOException {
        for ( Enumeration e = getMetadata().getFieldDescriptors(); e.hasMoreElements() ;) {
            FieldDescriptor field = (FieldDescriptor) e.nextElement();
            Object value = get(field);
            field.checkValidValue(this, value);
        }
    }

    /**
     * Provides a default string representation of the
     * record for a human-readable flat-file
     * persistence scheme.
     *
     * @return string of form field=value \n field=value \n\n
     */
    public String toString() {
        Enumeration fields = getMetadata().getFieldDescriptors();
        StringBuffer buf = new StringBuffer();
        buf.append(TYPE_KEY);
        buf.append((char) '=');
        buf.append(getType());
        buf.append((char) '\n');
        if (!isImmutable()) {
            buf.append(STATUS_KEY);
            buf.append((char) '=');
            buf.append("mutable\n");
        }
        else if (isStale()) {
            buf.append(STATUS_KEY);
            buf.append((char) '=');
            buf.append("stale\n");
        }
        while ( fields.hasMoreElements() ) {
            FieldDescriptor field = (FieldDescriptor) fields.nextElement();
            String value = null;
            try {
                Object o = get(field);
                if ( o != null )
                    value = field.valueToString(o);
            }
            catch ( IOException ioe ) {
                throw new AssertionFailedException("It should be impossible to get here.");
            }
            if (value == null || "".equals(value))
                continue;
            buf.append(field);
            buf.append((char) '=');
            buf.append(value);
            buf.append((char) '\n');
        }
        buf.append((char) '\n');
        String result = buf.toString();
        if (Util.hasNonAscii(result)) {
            result = Util.toByteEncodedString(result);
        }
        return result;
    }


    /**
     * Called before checking the fields.
     * Does nothing in the base implementation.
     * Overriding this method allows you to apply certain
     * (possibly ad hoc) fixups before doing any
     * validity checks. Basically, this gives a class a chance
     * to massage data that wouldn't pass the validity check.
     */
    protected void preCheck() throws IOException {}

    /**
     * Called after checking the fields.
     * Does nothing in the base implementaiton.
     * If you want more, possibly cross-field or cross-record
     * validity checks, override this method and put
     * the checks here. If the data flunks, throw
     * the appropriate exception.
     */
    protected void postCheck() throws IOException {}

    public void freeze() throws IOException {
        checkValidity();
        immutable = true;
    }

    void checkFieldCompatibility(FieldDescriptor field) throws InvalidFieldException {
        if (!field.isFieldOf(this)) {
            throw new InvalidFieldException(this, field.toString() +
                                            " is not a field of this record");
        }
    }

    /**
     * Set the values of the record's fields
     * from a key-value mapping of the fields
     * (or field names) to values.
     * @param table key-value string pairs to set values of fields.
     * @return a string representing the status info contained in the table if any,
     * possibly null.
     */
    String setValuesFromTable(Map table) throws IOException {
        if (isImmutable()) {
            throw new ImmutableDataException(this);
        }
        String typeInfo = (String) table.remove(TYPE_KEY);
        if (this.type == null) {
            Assert.assert(typeInfo != null, "We need type info here");
            clearFields();
        }
        String statusInfo = (String) table.remove(STATUS_KEY);
        RecordDescriptor metadata = getMetadata();
        for (Iterator it = table.keySet().iterator(); it.hasNext();) {
            String key = (String) it.next();
            String value = (String) table.get(key);
            FieldDescriptor field = metadata.getField(key);
            if (field == null) {
                Logger.log("Invalid key name (" + key + ") in record of type " + type + ".",
                           Logger.LOG_WARN);
            }
            else {
                set(field, field.getObjectFromString(this, value));
            }
        }
        return statusInfo;
    }

    public final Object get(String fieldname) {
        int index = getMetadata().indexOf(fieldname);
        if (index < 0)
            throw new InvalidFieldException(this, "unknown field " + fieldname);
        Object result = values[index];
        return getMetadata().getField(index).getCopy(result);
    }


    public final void set(String fieldname, Object value) {
        if (immutable)
            throw new ImmutableDataException(this);
        int index = getMetadata().indexOf(fieldname);
        FieldDescriptor field = getMetadata().getField(index);
        if (index <0) {
            throw new InvalidFieldException(this, "Unknown field: " +
                                            fieldname);
        }
        values[index] = field.getCopy(value);
    }

    public final Object get(FieldDescriptor field) {
        checkFieldCompatibility(field);
        return get(field.getName());
    }

    public final void set(FieldDescriptor field, Object value) {
        checkFieldCompatibility(field);
        set(field.getName(), value);
    }

    public final RecordDescriptor getMetadata() {
        if (descriptor == null) {
            if (type == null) {
                throw new AssertionFailedException("Cannot find metadata without type info. The most likely cause of this error is using new to create a new record instead of getExemplar() from a DataRegistry.");
            }
            try {
                RecordDescriptor rd = DataRegistryImpl.getDefaultRegistry().getMetadata(type);
                if (rd != null)
                    setMetadata(rd);
            }
            catch (IOException ioe) {
                throw new AssertionFailedException(ioe);
            }
            if (descriptor == null) {
                throw new AssertionFailedException("Could not get metadata " +
                                                   "for " + type);
            }
        }
        return descriptor;
    }

    public void readExternal(ObjectInput in)
    throws IOException, ClassNotFoundException
    {
        Map values = DataUtil.getFieldValueTable(this, in);
        String statusInfo = setValuesFromTable(values);
        if (getPrimaryKey() != null) {
            boolean validating = (in instanceof ValidatingRecordInput);
            if (statusInfo == null || "initialized".equals(statusInfo)) {
                preCheck();
                if (validating) //otherwise, we just assume the fields are okay.
                    checkFields();
                postCheck();
                this.immutable = true;
            }
        }
    }

    public void writeExternal(ObjectOutput out) throws IOException {
        out.write(toString().getBytes());
    }

    /**
     * create a clone.
     * A shallow copy of the values array
     * should be enough for immutable records
     * Note that a newly cloned object has its
     * immutable flag set to false,
     * so that it is in a modifiable state.
     */
    public Record getMutableCopy() {
        DefaultRecord result = null;
        try {
            result = (DefaultRecord) this.clone();
        }
        catch (CloneNotSupportedException cnse) {
            Assert.assert(false, "This should be impossible.");
        }
        // Put the record into the appropriate
        // state for a newly cloned record.
        result.immutable = result.stale = false;
        result.values = new Object[getMetadata().size()];
        System.arraycopy(this.values, 0, result.values, 0, values.length);
        return result;
    }

    protected final void set(int i, Object value) {
        if (immutable)
            throw new ImmutableDataException(this);
        FieldDescriptor field = getMetadata().getField(i);
        values[i] = field.getCopy(value);
    }

    protected final Object get(int i) {
        FieldDescriptor field = getMetadata().getField(i);
        return field.getCopy(values[i]);
    }

    // The only reason this method exists is for WebMacro to use.

    public Object get(Object o) {
        Object result = get(o.toString());
        return result == null ? "" : result;
    }


    /**
     * set the fields to their default values.
     */
    public void clearFields() {
        int index = 0;
        for (Enumeration e = getMetadata().getFieldDescriptors();
            e.hasMoreElements(); index++) {
            FieldDescriptor field = (FieldDescriptor) e.nextElement();
            set(index, field.defaultValue());
        }
    }

    public boolean equals(Object o) {
        if (o == this)
            return true;
        if (o == null || !o.getClass().equals(this.getClass()))
            return false;
        return toString().equals(o.toString());
    }

    public final void setMetadata(RecordDescriptor descriptor) {
        if (descriptor == this.descriptor)
            return;
        Assert.assert(this.descriptor == null);
        if (type != (null))
            Assert.assert(type.equals(descriptor.getType()));
        this.descriptor = descriptor;
        this.type = descriptor.getType();
        this.values = new Object[descriptor.size()];
    }

    public final String getType() {
        return type;
    }

    /**
     * The special reserved name of the type pseudo-key.
     */
    public static final String TYPE_KEY = "_meta";

    /**
     * The special reserved name of the status pseudo-key.
     */
    public static final String STATUS_KEY = "_status";
}
