/*
OREO Object-relational layer
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned 
by Reminders On-Line, Inc. It was released under the terms 
of the BSD-style open source license below. The terms of that 
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.oreo;

import com.revusky.oreo.util.*;
import java.io.*;

/**
 * An implementation of the ObjectInput interface
 * for slurping Records off a stream in their
 * default flat-file format, which is not
 * the standard Java serialised object format.
 * Objects are ASCII serialised in a format like this:
 *  _meta=reminder.1
 *  unique_id=1
 *  creation_date=Nov 17, 1999 8:03:15 AM
 *  name=vitamin
 *  ...
 * The stream will be a complete clean rewrite of all live persistent objects followed
 * by some appended delete and add records of changes since the last complete rewrite.
 * The append section could contain obsolete records, deleted and added repeatedly.
 * @author Jonathan Revusky jon@revusky.com
 */

public class DefaultRecordInputStream 
extends DataInputStream implements ObjectInput, ValidatingRecordInput {

    /**
     * Read the next record off the stream.
     * @return null if this is the end of the stream.
     */
    public Object readObject() throws IOException, ClassNotFoundException {
        try {
            String typeInfo = getTypeInfo();
            DefaultRecord rec = (DefaultRecord) getExemplar(typeInfo);
            if ( rec == null ) {
                Logger.log("Type " + typeInfo + " not known.", Logger.LOG_ERROR);
                throw new UnknownRecordTypeException(typeInfo);
            }
            rec.readExternal(this);
            if (inDeletion)
                rec = new AntiRecord(rec);
            return rec;
        }
        catch (EOFException eof) {
            return null;
        }
    }

    /**
     * Read the first line of the record data,
     * which tells us what type of record this is.
     * @return record type e.g. "reminder.1 
     */
    protected String getTypeInfo() throws IOException {
        inDeletion=false;
        String line = "";
        while (!line.startsWith("_meta=")) {
            if (line.startsWith(DELETION_MARKER))
                inDeletion = true;
            else if (line.startsWith(INSERTION_MARKER) || line.startsWith(UPDATE_MARKER))
                inDeletion = false;
            line = this.readLine();
            if ( line == null ) {
                throw new EOFException("End of input looking for next record");
            }
        }
        int equalsIndex = line.indexOf('=');
        return line.substring(1+equalsIndex);
    }

    private Record getExemplar(String type) throws IOException {
        DataRegistry registry = DataRegistryImpl.getDefaultRegistry();
        RecordDescriptor rd = registry.getMetadata(type);
        if (rd == null) {
            int periodIndex = type.indexOf('.'); // backward-compatibility kludge
            if ( periodIndex >=0 ) {
                type = type.substring(0, periodIndex);
                rd = registry.getMetadata(type);
            }
            if (rd == null)
                return null;
        }
        return rd.newRecord();
    }


    /**
     * Constructor, given stream
     * @param is the underlying stream to read persistent data from.
     */
    public DefaultRecordInputStream(InputStream is) {
        super(is);
    }

    /**
     * a convenience constructor that takes a File argument.
     * @param file   File from which to read the persistent data.
     * @exception IOException
     */
    public DefaultRecordInputStream(File file) throws IOException {
        this(new BufferedInputStream(new FileInputStream(file), 2048));
    }

    private boolean inDeletion;
    public static final String DELETION_MARKER="#!DELETION";
    public static final String INSERTION_MARKER="#!INSERTION";
    public static final String UPDATE_MARKER="#!UPDATE";
}

