/*
OREO Object-relational layer
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned
by Reminders On-Line, Inc. It was released under the terms
of the BSD-style open source license below. The terms of that
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.oreo;

import java.io.IOException;
import java.util.List;

/**
 * Interface implemented by objects that manage a collection
 * of records.
 *
 * @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
 */
public interface MutableDataSource extends DataSource {

    /**
     * Adds a new record to the managed pool.
     *
     * @param rec the record to add
     *
     * @exception DuplicateRecordException if another record already exists
     * with same primary key as 'rec'.
     *
     * @exception IOException if the record cannot be
     * initialized (i.e. has missing fields or invalid
     * field values, etc.), or in case of a low-level error.
     */
    void insert(Record rec) throws IOException;

    /**
     * Replaces an existing version of a record with a new updated version.
     *
     * Some implementations may use the oldRec parameter
     * to guarantee that there are no concurrency issues.
     * In such case, it would throw a ConcurrentModificationException
     *
     * @param oldRec the record to replace.
     * @param newRec the new record.
     *
     * @exception IOException thrown in case of any other database or
     * communication error.
     */
    void update(Record oldRec, Record newRec) throws IOException;

    /**
     *  Delete a record
     *
     *  @param the record to delete.
     */
    void delete(Record rec) throws IOException;

    /**
     *  Fetches a list of records matching filter.  If filter is null,
     *  returns all the records in the container.
     *
     *  @param filter the record filter, or null.
     *
     *  @return a List of all the records that are an instance of a
     *  given class.  If no records match the filter, an empty List.
     */
    List select(RecordFilter filter) throws IOException;

    /**
     * "play" an event
     * This will be mostly used in restarting an app
     * and reconstructing the state of the data from logs.
     * Note that data listeners are not notified.
     */

    void playEvent(DataEvent event) throws IOException;


   /**
     * Add a listener that gets notified if ever a record is inserted,
     * updated or deleted.
     *
     * @param dl the listener
     */
    void addDataListener(DataListener dl) throws IOException;


    /**
     * Remove a listener that gets notified if ever a record is inserted,
     * updated or deleted.
     *
     * @param dl the listener
     */
    void removeDataListener(DataListener dl) throws IOException;
}
