/*
OREO Object-relational layer
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned 
by Reminders On-Line, Inc. It was released under the terms 
of the BSD-style open source license below. The terms of that 
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.oreo;

import java.io.*;
import java.util.Locale;
import java.util.Date;

/**
 * An interface that describes the canonical
 * methods of a data record in the Oreo Object-Relational Layer.
 * <P>
 * Basically, a Record should have the semantics of
 * an immutable set of values, where the values follow
 * a scheme given by the metadata in the associated RecordDescriptor
 * instance that can be got at via the method getMetadata()
 * <P>
 * The immutability semantics have a nuance.
 * A record's fields can be modified via set()
 * if the record is in a mutable state. Once a record is <em>frozen</EM>
 * its values are immutable. This typically occurs when the record is added to a
 * <CODE>MutableDataSource</CODE> container.
 * <P>
 * In this library, the way to "modify" data is via the
 * getMutableCopy() method. This will give you a copy that
 * can be modified. When that mutable copy is added to the
 * MutableDataSource, the new version will replace the old
 * version. The newer records will be frozen, i.e. immutable.
 * Also, the older record will be marked stale, to indicate
 * to any thread or object holding onto a reference that it has been
 * superseded. 
 * <P>
 * The real virtue of this scheme is that when you are
 * working with a record that is immutable, you have
 * a pretty much iron-clad guarantee that all of the data in
 * the record is "valid" -- at least in the sense that it
 * matches the metadata description given by getMetadata().
 * You also have none of the typical synchronization worries,
 * where you have to think about the possibility that
 * another thread can change any values out from under you.
 *  <p>
 *  Each record has a primary key which may be set automaticaly by
 *  the MutableDataSource when the record first put under
 *  management using <code>insert()</code>.
 *  <p>
 * @see com.revusky.oreo.RecordDescriptor
 * @see com.revusky.oreo.DefaultRecord
 * @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
 */

public interface Record {
    /**
     * @return the name of this record type
     * subclasses must implement this. e.g. "reminder"
     */
    String getType();

    /**
     *  Retrieve the value of the record's primary key.
     *
     *  @return the value of this record's primary key, or null if it
     *  has not been set.
     */
    Object getPrimaryKey();

    /**
     *  set the value of this record's primary key.
     *  The record must be in a mutable state
     *  AND the primary key must be unset.
     */
    void setPrimaryKey(Object o);

    /**
     * create a clone.
     * A shallow copy of the record values
     * should be enough for immutable records
     * Note that a newly cloned object is
     * in a mutable state, so that you can modify it.
     */
    Record getMutableCopy();

    /**
      * Have the fields all been set?
      * Once a record is put in a 
      * DataSource, its fields are immutable.
      */
    boolean isImmutable();

    /**
     * Has this record been deleted or superseded in
     * the associated DataSource?
     */
    boolean isStale();

    /**
     * make this record immutable. 
     * This will typically only be called by a MutableDataSource object.
     */
    void freeze() throws IOException;

    /**
     * @return an object that describes the data held in this record.
     * @see com.revusky.oreo.RecordDescriptor
     */
    RecordDescriptor getMetadata();

    /**
     * Method to get the value of a field by name.
     * Concrete implementations may wrap this
     * with a higher-level get/set API.
     * @param fieldname
     * @throws InvalidFieldException if there is no field of that name.
     */
    Object get(String fieldname);

    /**
     * Low-level method to query the value of a field in a Record.
     * concrete implementations will probably wrap this.
     * @param field to query.
     * @return an Object wrapper around field value.
     */
    Object get(FieldDescriptor field);

    /**
     * Low-level method to set an individual field value
     * concrete implementations will probably wrap this.
     * @param field to set.
     * @param value Object wrapping the value
     * @throws InvalidFieldException if there is no field of that name.
     * @throws ImmutableDataException if this record
     * is immutable @see #getMutableCopy()
     */
    void set(FieldDescriptor field, Object value);


    /**
     * Method to set the value of a field by name.
     * Concrete implementations may wrap this
     * with a higher-level get/set API.
     * @param fieldname the name of the field to set.
     * @throws InvalidFieldException if there is no field of that name.
     * @throws ImmutableDataException if this record
     * is immutable @see #getMutableCopy()
     */
    void set(String fieldname, Object value);

    /**
     * reset the fields to their default state
     * The record must be in a mutable state.
     * @throws ImmutableDataException if this record
     * is immutable @see #getMutableCopy()
     */
    void clearFields();

    /**
     * Method only used internally.
     */
    void setMetadata(RecordDescriptor desc);
}

