/*
OREO Object-relational layer
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned 
by Reminders On-Line, Inc. It was released under the terms 
of the BSD-style open source license below. The terms of that 
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.oreo;

import java.io.*;
import com.revusky.oreo.util.*;

/**
 * An interface for objects that can filter a set of records
 * based on some criterion. More formally speaking: a unary predicate 
 * that operates on collections of records.
 * A RecordFilter has a method called accept() that allows
 * us to filter over a set of in-memory records. The 
 * sqlEquivalent() should return the SQL query
 * that would provide the same subset of records from an 
 * RDBMS with an SQL interface. It is the responsibility 
 * of the implementor of a subclass that the SQL statement
 * actually is equivalent to the accept() method. If not, 
 * all kinds of bugs and mayhem may result!
 * @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
 * @see com.revusky.oreo.predicates
 */

public interface RecordFilter extends Serializable {

    /**
     * Does this record meet the filter criterion?
     */
    boolean accept(Record rec);


    /**
     * says what kind of record type we are looking for, i.e.
     * the filter will only return results of a certain record type.
     * If this returns null, it means that we could have heterogeneous results.
     */

    String recType();



    /**
     * @return The SQL SELECT command that is the equivalent to this. A null
     * value would indicate that we don't know the SQL command.
     */
    String sqlEquivalent();
}
