/*
OREO Object-relational layer
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned
by Reminders On-Line, Inc. It was released under the terms
of the BSD-style open source license below. The terms of that
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.oreo.inmemory;

import com.revusky.oreo.*;
import com.revusky.oreo.util.*;
import java.io.*;
import java.util.*;

/**
 *  A simple implementation of the com.revusky.oreo.MutableDataSource
 *  interface. All records are kept in an flat array all in RAM and
 *  persisted to a flat-file. In configuring this file in XML,
 * you only have to specify 2 properties: STORE, which indicates
 * a file location to persist to. You can optionally indicate COMPACT_FREQUENCY,
 * which says how frequently to rewrite the entire store in a compact form.
 * @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
 */
public class InMemoryRecordSet extends AbstractMutableDataSource {
    /**
     *  The underlying container in which we store the records.
     */
    private Map container = new HashMap();

    /**
     *  An integer to keep track of what the next integer to auto-assign to
     *  a record is.
     */
    private int nextIntegerKey;

    public InMemoryRecordSet() throws IOException {
    }

    /**
     * Slurp the records into this InMemoryRecordSet object with a
     * stream-based idiom. Subclasses of DataException will be
     * caught and logged. Core Java IOExceptions will bubble up.
     * NB: This method is meant to be called at initialization.
     * The loading of the records does not go through the DataListener
     * mechanism.
     */
    public void loadRecords(ObjectInput input) throws IOException {
        Record rec = new DefaultRecord(); // just so rec !=null
        while (rec != null) {
            try {
                rec = (Record) input.readObject();
                if (rec != null)
                    load(rec);
            }
            catch (DataException mde) {
                Logger.log(mde.getMessage(), Logger.LOG_WARN);
            }
            catch (ClassNotFoundException cnfe) {
                Logger.log(cnfe.getMessage(), Logger.LOG_WARN);
            }
        }
    }

    public synchronized void update(Record oldRec, Record newRec) throws IOException {
        Assert.assert(!newRec.isImmutable(),
                      "Expecting a mutable record.");
        Assert.assert(!newRec.isStale(),
                      "This record has been deleted. " +
                      "You cannot add it to the data source.");
        Assert.assert(oldRec.isImmutable(), "You can only update a record that is already initialized.");
        if (oldRec.isStale()) {
            throw new ModifiedRecordException(newRec, "Attempt to update a version that is not the latest one.");
        }
        Object primaryKey = newRec.getPrimaryKey();
        Assert.assert(oldRec.getPrimaryKey().equals(primaryKey));
        // The record is frozen, and hence, validated and immutable.
        freeze(newRec);
        markStale(oldRec);
        load(newRec);
        fireDataEvent(new DataEvent(this,
                                    DataEvent.RECORD_UPDATED,
                                    newRec,
                                    oldRec));
    }

    public synchronized void insert(Record record) throws IOException {
        Object oldKey = record.getPrimaryKey();
        Object key = oldKey;
        if (key == null || "".equals(key))
            key = nextKey();
        if (container.get(key) != null)
            throw new DuplicateRecordException(record,
                                               "The primary key "
                                               + key
                                               + " is already used.");
        record.setPrimaryKey(key);
        try {
            freeze(record);
        }
        catch (IOException ioe) {
            record.setPrimaryKey(oldKey);
            throw ioe;
        }
        load(record);
        fireDataEvent(new DataEvent(
                                   this,
                                   DataEvent.RECORD_INSERTED,
                                   record,
                                   null));
    }

    public synchronized List select(RecordFilter filter) throws IOException {
        ArrayList result = new ArrayList();
        Iterator it = container.values().iterator();
        while (it.hasNext()) {
            Record rec = (Record) it.next();
            if (filter !=null && (!filter.accept(rec)))
                continue;
            result.add(rec);
        }
        result.trimToSize();
        return result;
    }

    /**
     *  Called by <code>loadRecords()</code>.
     *  <p>
     *  Unlike <code>insert()</code>, this method does not
     *  notify any listeners.
     *  <p>
     *  @param record the record to be added.
     */
    void load(Record record) throws IOException {
        freeze(record); // make sure the record is immutable after this point.
        Object key = record.getPrimaryKey();
        Record oldRec = (Record) container.get(key);
        if ( oldRec != record ) {
            if (oldRec != null) {
                markStale(oldRec);
                container.remove(key);
            }
            if (!(record instanceof AntiRecord)) {
                container.put(key, record);
                if (key instanceof Integer) {
                    int id = ((Integer) key).intValue();
                    if (id >= nextIntegerKey)
                        nextIntegerKey = id + 1;
                }
            }
        }
    }

    public synchronized Record get(String type, Object key) throws IOException {
        if (key == null)
            return null;
        Record rec = (Record) container.get(key);
        if (rec == null && (key instanceof String)) {
            Object normkey = normalizeKey(type, (String) key);
            if (!key.equals(normkey))
                rec = (Record) container.get(normkey);
        }
        if (type != null
            && rec!= null
            && !type.equals(rec.getType())) {
            rec = null;
        }
        return rec;
    }

    public void init(Properties props) throws IOException {
        super.init(props);
        String filename = props.getProperty("STORE");
        if (filename.startsWith("file:/"))
            filename = filename.substring("file:/".length());
        if (new File(filename).exists())
            loadRecordsFromFlatFile(filename);
        String pfs = props.getProperty("COMPACT_FREQUENCY");
        if (pfs == null)
            pfs = props.getProperty("PERSIST_FREQUENCY"); // a bit of backward compatibility.
        int persistFrequency = 1000;
        if (Util.hasContent(pfs)) {
            try {
                persistFrequency = Integer.parseInt(pfs);
            }
            catch (NumberFormatException nfe) {
                Logger.log("expecting number here, got: COMPACT_FREQUENCY=\"" + pfs + "\"");
            }
        }
        new CrudeDataPersister(this, filename, persistFrequency);
    }

    public synchronized List keys(String type) throws IOException {
        Set allKeys = container.keySet();
        if (!Util.hasContent(type))
            return new ArrayList(allKeys);
        ArrayList result = new ArrayList();
        for (Iterator it = allKeys.iterator(); it.hasNext();) {
            Object key = it.next();
            Record rec = (Record) container.get(key);
            if (rec != null && rec.getType().equals(type))
                result.add(key);
        }
        result.trimToSize();
        return result;
    }

    public synchronized void delete(Record record) throws IOException {
        markStale(record);
        container.remove(record.getPrimaryKey());
        fireDataEvent(new DataEvent(this,
                                    DataEvent.RECORD_DELETED,
                                    null,
                                    record));
    }

    public void close() throws IOException {}

    /**
     * Returns the next available primary key,
     * assuming that it is an integer.
     */
    protected Object nextKey() {
        return new Integer(nextIntegerKey++);
    }

    public void playEvent(DataEvent event) throws IOException {
        switch (event.type) {
            case DataEvent.RECORD_DELETED :
                container.remove(event.oldRecord.getPrimaryKey());
                break;
            case DataEvent.RECORD_UPDATED :
                container.remove(event.oldRecord.getPrimaryKey());
            case DataEvent.RECORD_INSERTED :
                load(event.newRecord);
                break;
            default :
                throw new AssertionFailedException("Do not know how to play this kind of event.");
        }
    }


    /**
     * Wrapper around loadRecords() to read records from a flat file.
     * Creates a DefaultRecordInputStream and invokes
     * the record manager's loadRecords() method.
     * Will try to pick it up from the .### temp file if that
     * is available and filename is not.
     * NB: This method is meant to be called at initialization.
     * The loading of the records does not go through the DataListener
     * mechanism.
     */
    protected void loadRecordsFromFlatFile(String filename) throws IOException {
        ObjectInput input = null;
        FileLock lock = null;
        File file = new File(filename);
        if (!file.exists()) {
            File tempFile = new File(filename + ".###");
            if ( !tempFile.exists() ) {
                // Just warn here.
                Logger.log("File " +
                           filename + " not found or accessible.",
                           Logger.LOG_WARN);
                return;
            }
            else {
                // If there is no file of this name, but there
                // is a filename.### we read from that.
                file = tempFile;
                Logger.log("File "
                           + filename
                           + " not found, using "
                           + tempFile,
                           Logger.LOG_WARN);
            }
        }

        try {
            lock = new FileLock(filename);
            input = new DefaultRecordInputStream(file);
            loadRecords(new DefaultRecordInputStream(file));
        }
        finally {
            if (lock != null)
                lock.dispose();
            if (input!=null)
                input.close();
        }
    }
}
