/*
OREO Object-relational layer
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned 
by Reminders On-Line, Inc. It was released under the terms 
of the BSD-style open source license below. The terms of that 
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.oreo.jdbc;

import java.io.*;
import java.sql.*;
import java.text.*;
import java.util.*;
import com.revusky.oreo.*;
import com.revusky.oreo.util.*;

/**
 * A holder for some JDBC/SQL-related convenience routines
 */

abstract public class SQLUtil {

    static String buildInsertQuery(Record rec) {
        StringBuffer buf = new StringBuffer("INSERT INTO ");
        RecordDescriptor metadata = rec.getMetadata();
        buf.append(metadata.getTableName());
        buf.append("(");
        boolean first = true;
        for (Enumeration e = metadata.getFieldDescriptors(); e.hasMoreElements();) {
            FieldDescriptor fd = (FieldDescriptor) e.nextElement();
            if (!first) {
                buf.append(", ");
            }
            buf.append(fd.getName());
            first = false;
        }
        buf.append(") VALUES (");
        first = true;
        for (Enumeration e = metadata.getFieldDescriptors(); e.hasMoreElements();) {
            FieldDescriptor fd = (FieldDescriptor) e.nextElement();
            if (!first) {
                buf.append(", ");
            }
            buf.append((char) '?');
            first = false;
        }
        buf.append(")");
        return buf.toString();
    }

    static String buildUpdateQuery(Record rec) {
        StringBuffer buf = new StringBuffer("UPDATE ");
        RecordDescriptor metadata = rec.getMetadata();
        FieldDescriptor primaryKeyField = metadata.getPrimaryKeyField();
        buf.append(metadata.getTableName());
        boolean first = true;
        for (Enumeration e = metadata.getFieldDescriptors(); e.hasMoreElements();) {
            FieldDescriptor fd = (FieldDescriptor) e.nextElement();
            if (fd == primaryKeyField)
                continue;
            if (first)
                buf.append(" SET ");
            else
                buf.append(",");
            buf.append(fd.getName());
            buf.append("= ?");
            buf.append((char) '?');
            first = false;
        }
        buf.append(" WHERE ");
        buf.append(primaryKeyField.getName());
        buf.append("= ?");
        return buf.toString();
    }


    static public void fillRecordFromResultSet(Record rec, ResultSet rs) throws SQLException {
        for (Enumeration enum = rec.getMetadata().getFieldDescriptors(); enum.hasMoreElements();) {
            FieldDescriptor field = (FieldDescriptor) enum.nextElement();
            Object val = rs.getObject(field.getName());
            rec.set(field, field.convertFromJDBC(val));
        }
    }

    /**
     * Creates a table in the database or tries to do so.
     */
    static public boolean createTable(Connection conn, RecordDescriptor metadata) throws SQLException {
        String tableName = metadata.getTableName();
        Statement stmt = conn.createStatement();
        StringBuffer buf = new StringBuffer("CREATE TABLE " + tableName + " ( ");
        FieldDescriptor primaryKeyField = metadata.getPrimaryKeyField();
        Enumeration fields = metadata.getFieldDescriptors();
        while (fields.hasMoreElements()) {
            FieldDescriptor field = (FieldDescriptor) fields.nextElement();
            boolean isPrimaryKey = (field == primaryKeyField);
            buf.append(field.getName());
            buf.append((char) ' ');
            buf.append(field.getSQLType());
            if (isPrimaryKey ||field.isRequired()) 
                buf.append(" NOT NULL");
            if (isPrimaryKey) 
                buf.append(" PRIMARY KEY");
            if (fields.hasMoreElements())
                buf.append((char) ',');
        }
        buf.append(")");
        String command = buf.toString();
        Logger.log("Creating table by executing SQL command: " + command, Logger.LOG_INFO);
        stmt.executeUpdate(command);
        stmt.close();
        return true;
    }

    /**
     * @return a List containing all the table names.
     * @param conn A connection to a database.
     */
    static public List getTableNames(Connection conn) throws SQLException {
        List result = new ArrayList();
        DatabaseMetaData meta = conn.getMetaData();
        ResultSet rs = meta.getTables(null, null,  null, new String[] {"TABLE"});
        while (rs.next()) {
            String tableName = rs.getString(3);
            if (tableName != null) {
                result.add(tableName.toLowerCase());
            }
        }
        return result;
    }

    static public int numRowsInTable(Connection conn, String tableName) {
        try {
            Statement st = conn.createStatement();
            ResultSet rs = st.executeQuery("SELECT COUNT(*) FROM " + tableName);
            rs.next();
            return rs.getInt(1);
        }
        catch (SQLException e) {
            return -1;
        }
    }
}
