/*
OREO Object-relational layer
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned 
by Reminders On-Line, Inc. It was released under the terms 
of the BSD-style open source license below. The terms of that 
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.oreo.metadata;

import com.revusky.oreo.*;
import com.revusky.oreo.util.*;
import java.io.IOException;
import java.io.Serializable;
import java.util.*;
import org.w3c.dom.*;
import org.xml.sax.SAXException;


/**
 * abstract base class for objects that represent
 * the metadata relating to a record field. This object knows
 * how to set itself up from a DOM element.
 * @author <a href="mailto:jon@revusky.com>Jonathan Revusky</a>
 */
abstract public class Field implements FieldDescriptor {

    String name;
    private boolean required;
    protected String sqlType;
    protected String defaultValueString;
    protected Properties properties = new Properties();
    transient private Object defaultValue;

    /**
     * A bitmask that determines how the string is "normalized".
     */
    int normalization;

    static final int LOWER_CASE = 1;
    static final int UPPER_CASE = 2;
    static final int CAPITALIZE = 4;
    static final int TRIM = 8;
    static final int TRIM_EXCESS_WHITESPACE = 16;

    // only relevant if maxLength is set.
    static final int TRUNCATE_SILENTLY = 32;

    // only relevant if minLength is set.
    static final int PAD_SILENTLY = 64;

    /**
     * The whitespace characters used when trimming
     * strings for normalization.
     */
    private String whitespacechars = " \t";

    public Field() {
    }

    /**
     * If you extend this class and override this method,
     * do make sure to call the super.init() at some point.
     */
    protected void init(Element element) throws SAXException {
        this.name = element.getAttribute("NAME");
        if (!Util.hasContent(name))
            throw new SAXException("A field must have a name specified");
        name = name.intern();
        this.required = DOMUtil.isAffirmative(element.getAttribute("REQUIRED"));
        String sqlType = element.getAttribute("SQL_TYPE");
        if (Util.hasContent(sqlType))
            this.sqlType = sqlType;
        defaultValueString = element.getAttribute("DEFAULT");
        try {
            processDefaultValue();
        }
        catch (Exception e) {
            throw new SAXException(e);
        }
        Iterator propertyElements = DOMUtil.getChildElementsByName(element, "PROPERTY");
        while (propertyElements.hasNext()) {
            Element e = (Element) propertyElements.next();
            String key = e.getAttribute("KEY");
            String value = e.getAttribute("VALUE");
            if (!Util.hasContent(key))
                throw new SAXException("Must specify KEY attribute for PROPERTY in DATASOURCE: " + name);
            if (value == null)
                throw new SAXException("Must specify VALUE attribute for PROPERTY " + key + " in DATASOURCE: " + name);
            if (properties.getProperty(key) != null)
                throw new SAXException("KEY attribute " + key + " can only be specified once in DATASOURCE: " + name);
            properties.setProperty(key, value);
        }
        processProperties();
        Iterator normalize = DOMUtil.getChildElementsByName(element, "NORMALIZE");
        while (normalize.hasNext()) {
            processNormalization((Element) normalize.next());
        }
    }

    protected void processDefaultValue() throws IOException {
        if (Util.hasContent(defaultValueString)) {
            setDefaultValue(getObjectFromString(null, defaultValueString));
        }
    }

    protected void processProperties() throws SAXException {
        String stype = properties.getProperty("SQL_TYPE");
        if (stype != null) {
            this.sqlType = stype;
        }
        String defValue = properties.getProperty("DEFAULT");
        if (defValue != null) {
            this.defaultValueString = defValue;
            try {
                processDefaultValue();
            }
            catch (Exception e) {
                throw new SAXException(e);
            }
        }
    }

    protected void processNormalization(Element elem) {
        String type = elem.getAttribute("TYPE");
        if ("LOWER_CASE".equals(type))
            normalization |= LOWER_CASE;
        else if ("UPPER_CASE".equals(type))
            normalization |= UPPER_CASE;
        else if ("CAPITALIZE".equals(type))
            normalization |= CAPITALIZE;
        else if ("TRIM".equals(type))
            normalization |= TRIM;
        else if ("TRIM_EXCESS_WHITE_SPACE".equals(type)) {
            normalization |= TRIM_EXCESS_WHITESPACE;
            String whitespacechars = properties.getProperty("WHITESPACECHARS");
            if (Util.hasContent(whitespacechars))
                this.whitespacechars = whitespacechars;
        }
        else if ("TRUNCATE_SILENTLY".equals(type)) {
            normalization |= TRUNCATE_SILENTLY;
        }
    }


    public Object getCopy(Object o) {
        return o;
    }

    public final String getName() {
        return name;
    }

    public String toString() {
        return name;
    }

    public int hashCode() {
        return name.hashCode();
    }

    public boolean isRequired() {
        return required;
    }

    public String valueToString(Object val) {
        if (val == null)
            val = "";
        return val.toString();
    }

    public Class getJavaType() {
        return objectClass;
    }

    public String getSQLType() {
        if (sqlType == null)
	    sqlType = "VARCHAR(254)"; // punt
        return sqlType;
    }

    public void checkValidValue(Record rec, Object o) throws IOException {
        if (o == null || "".equals(o)) {
            if (this.isRequired())
                throw new MangledDataException(rec,
                                               "Field " + name + " is required.");
            return;
        }
        if (!getJavaType().isInstance(o)) {
            String message = "Invalid type for field " 
                + name 
                + ": expecting " 
                + getJavaType().getName() 
                + ", found " 
                + o.getClass().getName();
            throw new InvalidDataException(rec, message);
        }
    }

    public final boolean isFieldOf(Record rec) {
        return rec.getMetadata().getField(name) == this;
    }

    public Object defaultValue() {
        if (defaultValue == null && Util.hasContent(defaultValueString)) {
            try {
                processDefaultValue();
            }
            catch (IOException ioe) {
                throw new AssertionFailedException(ioe);
            }
        }
        return defaultValue;
    }

    // FIXME

    public String getLocalizedName(Locale locale) {
        StringTokenizer st = new StringTokenizer(name, "_", true);
        StringBuffer result = new StringBuffer();
        while (st.hasMoreTokens()) {
            String tok = st.nextToken();
            if (tok.equals("_"))
                result.append(" ");
            else
                result.append(capitalizeFirstLetter(tok));
        }
        return result.toString();
    }

    protected void setDefaultValue(Object o) {
        try {
            checkValidValue(null, o);
            this.defaultValue = o;
        }
        catch (IOException ioe) {
            Assert.rethrow(ioe);
        }
    }

    private String capitalizeFirstLetter(String s) {
        StringBuffer buf = new StringBuffer();
        StringTokenizer st = new StringTokenizer(s, whitespacechars);
        while (st.hasMoreTokens()) {
            String tok = st.nextToken();
            int len = tok.length();
            String firstChar = (len >0) ? tok.substring(0, 1) : "";
            String trailing = (len > 1) ? tok.substring(1) : "";
            buf.append(firstChar.toUpperCase());
            buf.append(trailing.toLowerCase());
            if (st.hasMoreTokens()) {
                buf.append((char) ' ');
            }
        }
        return buf.toString();
    }

    public String normalize(String s) {
        if ((normalization & LOWER_CASE) != 0)
            s = s.toLowerCase();
        if ((normalization & UPPER_CASE) != 0)
            s = s.toUpperCase();
        if ((normalization & CAPITALIZE) != 0) {
            s = capitalizeFirstLetter(s);
        }
        if ((normalization & TRIM) != 0)
            s = s.trim();
        if ((normalization & TRIM_EXCESS_WHITESPACE) != 0) {
            StringTokenizer st = new StringTokenizer(s, whitespacechars);
            StringBuffer buf = new StringBuffer();
            while (st.hasMoreTokens()) {
                buf.append(st.nextToken());
                buf.append((char) ' ');
            }
            s = buf.toString().trim();
        }
        return s;
    }

    public Object convertFromJDBC(Object o) {
        return o;
    }

    public String getProperty(String propertyName) {
        return properties.getProperty(propertyName);
    }


    static final Class objectClass = Object.class;
}
