/*
OREO Object-relational layer
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned
by Reminders On-Line, Inc. It was released under the terms
of the BSD-style open source license below. The terms of that
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.oreo.metadata;

import com.revusky.oreo.*;
import com.revusky.oreo.util.*;
import java.util.*;
import java.io.*;
import org.w3c.dom.*;
import org.xml.sax.SAXException;

/**
 * An implementation of the RecordDescriptor interface.
 * This class knows how to instantiate itself from a DOM Element
 * @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
 */
public class RecordMetadata implements RecordDescriptor {

    private Field[] fields;
    // Lookup for fieldname aliases.
    private HashMap aliases;
    private Class recordClass;
    private Class presentationWrapperClass;
//    private Field primaryKeyField;
    transient private Record exemplar;
    private String type;
    private String tableName;
    private Properties properties = new Properties();


    public RecordMetadata() {
    }

    /**
     * @param the name of the field or an alias to the name
     */
    public FieldDescriptor getField(String name) {
        for (int i = 0; i<fields.length; i++) {
            Field field = fields[i];
            if (field.name == name)
                return field;
        }
        for (int i = 0; i<fields.length; i++) {
            Field field = fields[i];
            if (field.name.equals(name))
                return field;
        }
        if (aliases !=null)
            return(FieldDescriptor) aliases.get(name);
        return null;
    }

    public FieldDescriptor getField(int i) {
        return fields[i];
    }

    public Enumeration getFieldDescriptors() {
        return Util.arrayAsEnumeration(fields);
    }

    /**
     * fill in internal state from a DOM element
     * @param element A W3C DOM API object that represents
     * the XML fragment that we are processing.
     * @param recordData a dictionary of metadata objects that have
     * been processed so far.
     * @param cloader The servlet classloader if it is available, or null
     * @throws SAXException This implies that there is something invalid
     * in the XML.
     */

    protected void init(Element element, Map recordData, ClassLoader cloader) throws SAXException {
        String classname = element.getAttribute("CLASS");
        if (!Util.hasContent(classname))
            classname = "com.revusky.oreo.DefaultRecord";
        try {
            this.recordClass = Util.findClass(classname, cloader);
        }
        catch (ClassNotFoundException e) {
            throw new SAXException("Could not find class " + classname);
        }
        type = element.getAttribute("TYPE").intern();
        String primaryKeyName = element.getAttribute("PRIMARY_KEY");
        processProperties(element);
        processBaseRecords(element, recordData);
        processPresentationWrapper(element, cloader);
        processFields(element, cloader);
        if (Util.hasContent(primaryKeyName)) {
            int pkIndex = indexOf(primaryKeyName);
            if (pkIndex < 0) {
                throw new SAXException("primary key named: "
                                       + primaryKeyName
                                       + " is not a field of this record.");

            }
            if (pkIndex >0) {
                // make sure that the first field is the primary key field.
                Field pkField = fields[pkIndex];
                fields[pkIndex] = fields[0];
                fields[0] = pkField;
            }

        }
        String tableName = element.getAttribute("TABLE_NAME");
        String tableName2 = properties.getProperty("TABLE_NAME");
        if (Util.hasContent(tableName) && Util.hasContent(tableName2))
            tableName = tableName2;
        if (Util.hasContent(tableName))
            this.tableName = tableName;
    }

    protected void processPresentationWrapper(Element element, ClassLoader cloader) throws SAXException
    {
        String classname  = element.getAttribute("PRESENTATION_CLASS");
        if (Util.hasContent(classname)) {
            try {
                presentationWrapperClass = Util.findClass(classname, cloader);
            }
            catch (Exception e) {
                throw new SAXException(e);
            }
        }
    }

    protected void processProperties(Element element) throws SAXException {
        NodeList children = element.getChildNodes();
        for (int i = 0; i<children.getLength(); i++) {
            Node child = children.item(i);
            if (!(child instanceof Element))
                continue;
            Element e = (Element) child;
            if (e.getTagName().equals("PROPERTY")) {
                String key = e.getAttribute("KEY");
                String value = e.getAttribute("VALUE");
                if (!Util.hasContent(key))
                    throw new SAXException("Must specify KEY attribute for PROPERTY in RECORD: " + type);
                if (value == null)
                    throw new SAXException("Must specify VALUE attribute for PROPERTY " + key + " in RECORD: " + type);
                if (properties.getProperty(key) != null)
                    throw new SAXException("KEY attribute " + key + " can only be specified once in RECORD: " + type);
                properties.setProperty(key, value);
            }
        }
    }


    /**
     * add the fields of the base record definition
     */
    private void processBaseRecords(Element element, Map recordData) throws SAXException {
        Iterator recs = DOMUtil.getChildElementsByName(element, "BASEREC");
        ArrayList baseFields = new ArrayList();
        while (recs.hasNext()) {
            Element baseRecElement = (Element) recs.next();
            String id = baseRecElement.getAttribute("NAME");
            RecordMetadata baseRec = (RecordMetadata) recordData.get(id);
            if (baseRec == null) {
                throw new SAXException("No record found of type =" + id);
            }
            for (int j = 0; j<baseRec.fields.length; j++) {
                Field f = baseRec.fields[j];
                if (!baseFields.contains(f))
                    baseFields.add(f);
            }
        }
        fields = new Field[baseFields.size()];
        fields = (Field[]) baseFields.toArray(fields);
    }


    /**
      * Fills in the fields array based on the FIELD subelements
      * in the RECORD element passed in.
      */
    private void processFields(Element element, ClassLoader cloader) throws SAXException {
        Iterator fieldNodes = DOMUtil.getChildElementsByName(element, "FIELD");
        ArrayList fieldList = new ArrayList();
        while (fieldNodes.hasNext()) {
            Element fieldElement = (Element) fieldNodes.next();
            String className = fieldElement.getAttribute("CLASS");
            try {
                Class cl = Util.findClass(className, cloader);
                Field field = (Field) cl.newInstance();
                field.init(fieldElement);
                String fieldName = field.getName();
                // Now check uniqueness of field name.
                for (Iterator it = fieldList.iterator(); it.hasNext();) {
                    Field f = (Field) it.next();
                    if (f.getName().equals(fieldName))
                        throw new SAXException("Field name " + fieldName
                                               + " was duplicated. Field names must be unique.");

                }
                fieldList.add(field);
                int j = fields.length;
                while (j>0) {
                    if (fields[--j].getName().equals(fieldName)) {
                        fieldList.remove(field);
                        fields[j] = field;
                    }
                }
                processAliases(fieldElement, field);
            }
            catch (Exception e) {
                if (e instanceof SAXException)
                    throw (SAXException) e;
                else // have to wrap it.
                    throw new SAXException(e);
            }
        }
        if (fields == null || fields.length == 0) {
            fields = new Field[fieldList.size()];
            fieldList.toArray(fields);
        }
        else {
            int numBaseFields = fields.length;
            int numNewFields = fieldList.size();
            Field[] ff = new Field[numBaseFields + numNewFields];
            System.arraycopy(fields, 0, ff, 0, numBaseFields);
            for (int i = 0; i<numNewFields; i++) {
                ff[numBaseFields + i] = (Field) fieldList.get(i);
            }
            this.fields = ff;
        }
    }

    public int size() {
        return fields.length;
    }

    /**
     * @return the index of the field of a given name, or -1 if it does not exist.
     * @param fieldname the name of the field or an alias
     */
    public int indexOf(String fieldname) {
        for (int i = 0; i < fields.length; i++) {
            if (fields[i].getName() == fieldname)
                return i;
        }
        for (int i = 0; i < fields.length; i++) {
            if (fields[i].getName().equals(fieldname))
                return i;
        }
        FieldDescriptor field = null;
        if (aliases != null) {
            field = (FieldDescriptor) aliases.get(fieldname);
        }
        if (field != null) {
            return indexOf(field.getName());
        }
        return -1;
    }

    private void processAliases(Element fieldElement, Field field) throws SAXException {
        Iterator aliasNodes = DOMUtil.getChildElementsByName(fieldElement, "ALIAS");
        while (aliasNodes.hasNext()) {
            Element alias = (Element) aliasNodes.next();
            String aliasName = DOMUtil.getText(alias);
            if (aliases != null && aliases.containsKey(aliasName)) {
                throw new SAXException("alias " + aliasName + " already defined.");
            }
            if (aliasName.equalsIgnoreCase(field.getName())) {
                throw new SAXException("alias " + aliasName + " must be different from the canonical field name.");
            }
            if (aliases == null)
                aliases = new HashMap();
            aliases.put(aliasName, field);
        }
    }

    public Class getPresentationWrapperClass() {
        return presentationWrapperClass;
    }

    public FieldDescriptor getPrimaryKeyField() {
        return fields[0];
    }

    public Class getRecordClass() {
        return recordClass;
    }

    public Record newRecord() {
        if (exemplar == null) {
            synchronized (this) {
                try {
                    exemplar = (Record) recordClass.newInstance();
                }
                catch (IllegalAccessException iae) {
                    throw new AssertionFailedException("Problem instantiating class: "
                                                       + recordClass.getName()
                                                       + "\nNB. A record implementation "
                                                       + "must be public and have a "
                                                       + "public no-args constructor");

                }
                catch (InstantiationException ie) {
                    throw new AssertionFailedException("Problem instantiating class: "
                                                       + recordClass.getName()
                                                       + "\nNB. Cannot instantiate "
                                                       + " an abstract class or interface.");
                }
                exemplar.setMetadata(this);
            }
        }
        Record result = exemplar.getMutableCopy();
        result.clearFields();
        return result;
    }

    public String getType() {
        return type;
    }

    public String getTableName() {
        return (tableName == null) ? getType() : tableName;
    }

    public String getProperty(String propertyName) {
        return properties.getProperty(propertyName);
    }
}
