/*
OREO Object-relational layer
Copyright (c) 2001, Jonathan Revusky, All rights reserved.
The original version of this source code is copyrighted and owned 
by Reminders On-Line, Inc. It was released under the terms 
of the BSD-style open source license below. The terms of that 
license also apply to the modified version here.
*/

/*
Copyright (c) 2000, Reminders On-Line, Inc. (ROL)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of the Reminders On-Line nor the names of its contributors
may be used to endorse or promote products derived from this software
without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.oreo.metadata;

import com.revusky.oreo.*;
import com.revusky.oreo.util.*;
import org.w3c.dom.*;
import gnu.regexp.*;
import java.util.*;
import java.io.IOException;
import java.util.StringTokenizer;
import org.xml.sax.SAXException;


/**
 * Field metadata that represents a field with a string value.
 * The string may be "normalized" to all caps, all lower case, etc.
 * Also, a regular expression pattern may be specified to validate
 * against.
 * @author <a href="mailto:jon@revusky.com">Jonathan Revusky</a>
 */
public class StringField extends Field {

    // The min and max lengths of the string.
    // Values of zero imply no minimum or maximum
    int minLength, maxLength;



    /**
     * The regexp that expresses a constraint
     * that the field must satisfy.
     */
    protected transient RE regexp;

    protected String regexpString;

    public StringField() {
    }

    protected void init(Element element) throws SAXException {
        super.init(element);
        Iterator expressions = DOMUtil.getChildElementsByName(element, "REGEXP");
        while (expressions.hasNext()) {
            processRegexp((Element) expressions.next());
        }
        if (!Util.hasContent(sqlType)) {
            if (maxLength >0) {
                sqlType =  "CHAR(" + maxLength + ")";
                if (minLength < maxLength) {
                    sqlType = "VAR" + sqlType;
                }
            }
        }
    }

    protected void processProperties() throws SAXException {
        super.processProperties();
        String smin = properties.getProperty("MIN_LENGTH");
        String smax = properties.getProperty("MAX_LENGTH");
        try {
            if (smin != null)
                minLength = Integer.parseInt(smin);
            if (smax != null)
                maxLength = Integer.parseInt(smax);
            if (minLength<0)
                throw new SAXException("MIN_LENGTH cannot be negative.");
            if (maxLength<0)
                throw new SAXException("MAX_LENGTH cannot be negative.");
        }
        catch (NumberFormatException nfe) {
            throw new SAXException("Expected number here.");
        }
        if ((minLength > 0) && (maxLength >0) && (minLength > maxLength)) {
            throw new SAXException("Minimum length cannot be greater than maximum length.");
        }
    }

    public Class getJavaType() {
        return stringClass;
    }

    public Object getObjectFromString(Record rec, String s) throws MangledDataException {
        return decodeMultiLineString(s);
    }

    public String normalize(String s) {
        s = super.normalize(s);
        if ((maxLength >0) && (normalization & TRUNCATE_SILENTLY) !=0) {
            if (s.length() > maxLength)
                s = s.substring(0, maxLength);
        }
        if (minLength >0 && (normalization & PAD_SILENTLY) != 0) {
            int diff = minLength - s.length();
            if (diff > 0) {
                StringBuffer buf = new StringBuffer(s);
                for (int i = 0; i< diff; i++)
                    buf.append((char) ' ');
                s = buf.toString();
            }
        }
        return s;
    }

    public String valueToString(Object o) {
        if (o == null)
            return null;
        String s = (String) o;
        if (s.indexOf('\\') >=0 || s.indexOf('\n') >=0 || s.indexOf('\r') >=0) {
            s = encodeMultiLineString(s);
        }
        return s;
    }

    public void checkValidValue(Record rec, Object value) throws IOException
    {
        super.checkValidValue(rec, value);
        String s = (String) value;
        if (!Util.hasContent(s)) // nothing to check.
            return;
        if (minLength > 0) {
            if (s.length() < minLength)
                throw new MangledDataException("String " + s + " should be at least " + minLength + " characters.");
        }
        if (minLength > 0) {
            if (s.length() > maxLength)
                throw new MangledDataException("String " + s + " should be at most  " + maxLength + " characters.");
        }
        if (regexp != null) {
            if (!regexp.isMatch(value)) {
                throw new MangledDataException(rec, "String " + value
                                               + " is not a valid value for field " + this.getName()
                                               + ". It does not match the pattern: " + regexp);
            }
        }
    }


    private void processRegexp(Element elem) throws SAXException {
        regexpString = elem.getAttribute("EXPRESSION");
        try {
            this.regexp = new RE(regexpString);
        }
        catch (Exception e) {
            throw new SAXException(e);
        }
    }

    /**
     * If the regexp string is defined, reconstitute the
     * regexp object (since gnu.regexp.RE is not serializable)
     */
    private void readObject(java.io.ObjectInputStream s)
    throws IOException, ClassNotFoundException
    {
        s.defaultReadObject();
        try {
            if (regexpString != null)
                this.regexp = new RE(regexpString);
        }
        catch (Exception e) {
            Assert.rethrow(e);
        }
    }

    /**
     * routine that encodes a string so that it can be all on one
     * line in a key=value storage scheme.
     */
    protected String encodeMultiLineString(String input) {
        StringBuffer buf = new StringBuffer();
        String specialChars = "=\r\n\\";
        StringTokenizer st = new StringTokenizer(input, specialChars, true);
        String token="", prevToken;
        while ( st.hasMoreTokens() ) {
            prevToken = token;
            token = st.nextToken();
            if (token.length() == input.length())
                return input;
            char c = token.charAt(0);
            if ((token.length() == 1) && (specialChars.indexOf(c) >=0)) {
                switch (c) {
                    case '\r' : 
                        buf.append("\\n");
                        break;
                    case '\n' :
                        if (!prevToken.equals("\r"))
                            buf.append("\\n");
                        break;
                    case '\\' :
                        buf.append("\\\\");
                        break;
                    case '=' :
                        buf.append("\\=");
                        break;
                }
            }
            else
                buf.append(token);
        }
        return buf.toString();
    }

    /**
     * Deal with a string with embedded \n literals for newlines.
     */
    protected String decodeMultiLineString(String input) {
        if ( input.indexOf('\\') <0 ) return input;
        StringBuffer buf = new StringBuffer();
        int i=0;
        while ( i<input.length() ) {
            char c = input.charAt(i++);
            if ( c != '\\' ) 
                buf.append(c);
            else {
                if (i < input.length()) {
                    c = input.charAt(i++);
                }
                if (c == 'n')
                    buf.append('\n');
                else if ( c== '\\' || c== '=')
                    buf.append(c);
                else {
                    buf.append((char) '\\');
                    buf.append(c);
                }
            }
        }
        return buf.toString();
    }

    public Object convertFromJDBC(Object o) {
        if (o == null)
            return null;
        if (!(o instanceof String)) {
            if (byteArrayClass.isInstance(o)) {
                o = new String((byte[]) o);
            }
            else if (charArrayClass.isInstance(o)) {
                o = new String((char[])  o);
            }
        }
        if ((normalization & TRIM) != 0)
            return o.toString().trim();
        else
            return o.toString();
    }

    public Object getCopy(Object o) {
        return (o == null) ? "" : o;
    }

    static final Class stringClass = String.class;
    static final Class byteArrayClass = ((Object) new byte[0]).getClass();
    static final Class charArrayClass = ((Object) new char[0]).getClass();
}

