/*
OREO Object-relational layer
Copyright (c) 2001, Jonathan Revusky
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice, 
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation 
and/or other materials provided with the distribution.

The name of Jonathan Revusky may not be used to endorse or promote 
products derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE. 
*/

package com.revusky.oreo.predicates;

import com.revusky.oreo.*;
import com.revusky.oreo.util.Assert;

/**
 * A Predicate that represents the AND, OR or XOR condition
 * on two other predicates.
 */

public class CompoundPredicate extends Predicate {

    public static final int UNION = 1;
    public static final int INTERSECTION = 2;

    final Predicate lhs, rhs;
    final int operation;

    protected CompoundPredicate(int operation, Predicate lhs, Predicate rhs) {
        this.lhs = lhs;
        this.rhs = rhs;
        this.operation = operation;
        Assert.assert(operation == UNION || operation == INTERSECTION);
        Assert.assert(lhs.recType().equals(rhs.recType()));
    }

    public String sqlEquivalent() {
        String conjunction = (operation == UNION) ? ") OR (" : ") AND (";
        if (operation == UNION) {
        }
        String query1 = lhs.sqlEquivalent();
        String query2 = rhs.sqlEquivalent();
        if (query1 == null || query2 == null)
            return null;
        if (query1.equals(query2))
            return query1;
        String from = SQLQueryUtil.from(query1);
        String where1 = SQLQueryUtil.where(query1);
        String where2 = SQLQueryUtil.where(query2);
        if (where1 == null)
            return query2;
        if (where2 == null)
            return query1;
        return "SELECT * FROM " 
            + from
            + " WHERE (" 
            + where1
            + conjunction
            + where2
            + ")";
    }

    public boolean accept(Record rec) {
        switch (operation) {
            case UNION:
                return lhs.accept(rec) || rhs.accept(rec);
            case INTERSECTION :
                return lhs.accept(rec) && rhs.accept(rec);
        }
        Assert.assert(false);
        return false;
    }

    public Predicate getComplement() {
        if (complement == null) {
            switch (operation) {
                case UNION:
                    complement = new CompoundPredicate(INTERSECTION, lhs.getComplement(), rhs.getComplement());
                    complement.complement = this;
                    break;
                case INTERSECTION :
                    complement = new CompoundPredicate(UNION, lhs.getComplement(), rhs.getComplement());
                    complement.complement = this;
                    break;
            }
        }
        return complement;
    }

    public String recType() {
        return lhs.recType();
    }
}
