/*
OREO Object-relational layer
Copyright (c) 2001, Jonathan Revusky
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice, 
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation 
and/or other materials provided with the distribution.

The name of Jonathan Revusky may not be used to endorse or promote 
products derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE. 
*/

package com.revusky.oreo.predicates;

import com.revusky.oreo.*;
import com.revusky.oreo.util.Assert;
import java.util.*;

/**
 * Holder for factory methods for predicate objects.
 * @see com.revusky.oreo.RecordFilter
 */

public class Predicates {

    /**
     * A trivial RecordFilter that simply accepts anything.
     */
    static public final RecordFilter NULL_FILTER = new TypeFilter(null);

    /**
    * this is equivalent to <CODE>getEqualityPredicate(recType, fieldName, comparisonValue, false)</CODE>
     */
    static public Predicate getEqualityPredicate(String recType, String fieldName, Object comparisonValue) {
        return getEqualityPredicate(recType, fieldName, comparisonValue, false);
    }

    /**
     * @param recType The type of record we are interested in.
     * @param fieldName The name of the field of the record we are filtering on.
     * @param comparisonValue The value to compare that field's value with.
    *  @param caseInsensitive whether to perform a case-insensitive match
     * @return A predicate that represents all records of a given type whose given field 
     * has the given a value.
    * <BR>If caseInsensitive is true, the comparisonValue must be a string (of course!)
     */
    static public Predicate getEqualityPredicate(String recType, String fieldName, Object comparisonValue, boolean caseInsensitive) {
        Assert.assert(!caseInsensitive || (comparisonValue instanceof String));
        int comparisonType = SimplePredicate.EQUALS;
        if (comparisonValue == null)
            comparisonType = SimplePredicate.IS_NULL;
        return new SimplePredicate(comparisonType, recType, fieldName, comparisonValue, null, caseInsensitive);
    }

    /**
     * @param recType The type of record we are interested in.
     * @param fieldName The name of the field of the record we are filtering on.
     * @param comparisonValue The value to compare that field's value with.
     * @param caseInsensitive whether to do a case-insensitive match.
     * @return A predicate that represents all records of a given type whose given field 
     * is <EM>not</EM> of the given value.
     * <BR>If caseInsensitive is true, the comparisonValue must be a string (of course!)
     */
    static public Predicate getInequalityPredicate(String recType, String fieldName, Object comparisonValue, boolean caseInsensitive) {
        Assert.assert(!caseInsensitive || (comparisonValue instanceof String));
        int comparisonType = SimplePredicate.NOT_EQUALS;
        if (comparisonValue == null)
            comparisonType = SimplePredicate.NOT_NULL;
        return new SimplePredicate(comparisonType, recType, fieldName, comparisonValue, null, caseInsensitive);
    }

    /**
    * this is equivalent to <CODE>getInequalityPredicate(recType, fieldName, comparisonValue, false)</CODE>
     */
    static public Predicate getInequalityPredicate(String recType, String fieldName, Object comparisonValue) {
        return getInequalityPredicate(recType, fieldName, comparisonValue, false);
    }

    /**
     * @param recType The type of record we are interested in.
     * @param fieldName The name of the field of the record we are filtering on.
     * @param comparisonValue The value to compare that field's value with.
     * @return A predicate that represents all records of a given type whose given field 
     * is <EM>less than</EM> the given value.
     */
    static public Predicate getLessThanPredicate(String recType, String fieldName, Object comparisonValue) {
        return getLessThanPredicate(recType, fieldName, comparisonValue, null);
    }

    /**
     * @param recType The type of record we are interested in.
     * @param fieldName The name of the field of the record we are filtering on.
     * @param comparisonValue The value to compare that field's value with.
     * @param comparator A custom comparator to use for comparisons
     * @return A predicate that represents all records of a given type whose given field 
     * is <EM>less than</EM> the given value.
     */
    static public Predicate getLessThanPredicate(String recType, String fieldName, Object comparisonValue, Comparator comparator) {
        return new SimplePredicate(SimplePredicate.LESS_THAN, recType, fieldName, comparisonValue, comparator);
    }

    /**
     * @param recType The type of record we are interested in.
     * @param fieldName The name of the field of the record we are filtering on.
     * @param comparisonValue The value to compare that field's value with.
     * @return A predicate that represents all records of a given type whose given field 
     * is <EM>greater than</EM> the given value.
     */
    static public Predicate getGreaterThanPredicate(String recType, String fieldName, Object comparisonValue) {
        return getGreaterThanPredicate(recType, fieldName, comparisonValue, null);
    }

    /**
     * @param recType The type of record we are interested in.
     * @param fieldName The name of the field of the record we are filtering on.
     * @param comparisonValue The value to compare that field's value with.
     * @param comparator A custom comparator to use for comparisons
     * @return A predicate that represents all records of a given type whose given field 
     * is <EM>greater than</EM> the given value.
     */
    static public Predicate getGreaterThanPredicate(String recType, String fieldName, Object comparisonValue, Comparator comparator) {
        return new SimplePredicate(SimplePredicate.GREATER_THAN, recType, fieldName, comparisonValue, comparator);
    }

    /**
     * @param recType The type of record we are interested in.
     * @param fieldName The name of the field of the record we are filtering on.
     * @param comparisonValue The value to compare that field's value with.
     * @return A predicate that represents all records of a given type whose given field 
     * is <EM>less than or equal to</EM> the given value.
     */
    static public Predicate getLessThanEqualsPredicate(String recType, String fieldName, Object comparisonValue) {
        return getLessThanEqualsPredicate(recType, fieldName, comparisonValue, null);
    }

    /**
     * @param recType The type of record we are interested in.
     * @param fieldName The name of the field of the record we are filtering on.
     * @param comparisonValue The value to compare that field's value with.
     * @param comparator A custom comparator to use for comparisons
     * @return A predicate that represents all records of a given type whose given field 
     * is <EM>less than or equal to</EM> the given value.
     */
    static public Predicate getLessThanEqualsPredicate(String recType, String fieldName, Object comparisonValue, Comparator comparator) {
        return new SimplePredicate(SimplePredicate.LESS_THAN_EQUALS, recType, fieldName, comparisonValue, comparator);
    }

    /**
     * @param recType The type of record we are interested in.
     * @param fieldName The name of the field of the record we are filtering on.
     * @param comparisonValue The value to compare that field's value with.
     * @return A predicate that represents all records of a given type whose given field 
     * is <EM>greater than or equal to</EM> the given value.
     */
    static public Predicate getGreaterThanEqualsPredicate(String recType, String fieldName, Object comparisonValue) {
        return getGreaterThanEqualsPredicate(recType, fieldName, comparisonValue, null);
    }

    /**
     * @param recType The type of record we are interested in.
     * @param fieldName The name of the field of the record we are filtering on.
     * @param comparisonValue The value to compare that field's value with.
     * @param comparator A custom comparator to use for comparisons
     * @return A predicate that represents all records of a given type whose given field 
     * is <EM>greater than or equal to</EM> the given value.
     */
    static public Predicate getGreaterThanEqualsPredicate(String recType, String fieldName, Object comparisonValue, Comparator comparator) {
        return new SimplePredicate(SimplePredicate.GREATER_THAN_EQUALS, recType, fieldName, comparisonValue, comparator);
    }

    /**
     * @param recType The type of record we are interested in.
     * @param fieldName The name of the field of the record we are filtering on.
     * @param comparisonValue The string to compare with
     * @param caseInsensitive whether to do a case-insensitive match.
     * @return A predicate that represents all records of a given type whose given field 
     * is <EM>greater than or equal to</EM> the given value.
     */
    static public Predicate getLikePredicate(String recType, String fieldName, String comparisonValue, boolean caseInsensitive) {
        return new SimplePredicate(SimplePredicate.LIKE, recType, fieldName, comparisonValue, null, caseInsensitive);
    }

    /**
     * @param recType The type of record we are interested in.
     * @param fieldName The name of the field of the record we are filtering on.
     * @param comparisonValue The string to compare with
     * @param caseInsensitive whether to do a case-insensitive match.
    * @return A predicate that represents all records of a given type whose given field 
     * is <EM>greater than or equal to</EM> the given value.
     */
    static public Predicate getNotLikePredicate(String recType, String fieldName, String comparisonValue, boolean caseInsensitive) {
        return new SimplePredicate(SimplePredicate.NOT_LIKE, recType, fieldName, comparisonValue, null, caseInsensitive);
    }

    /**
     * @param lhs left-hand side
     * @param  rhs right-hand side
     * @return A predicate that represents the intersection of the two conditions. A logical AND
     */
    static public Predicate getIntersection(Predicate lhs, Predicate rhs) {
        return new CompoundPredicate(CompoundPredicate.INTERSECTION, lhs, rhs);
    }

    /**
     * @param lhs left-hand side
     * @param  rhs right-hand side
     * @return A predicate that represents the union of the two conditions. A logical OR
     */
    static public Predicate getUnion(Predicate lhs, Predicate rhs) {
        return new CompoundPredicate(CompoundPredicate.UNION, lhs, rhs);
    }

    // static HashMap predicates = new HashMap();
}
