/*
OREO Object-relational layer
Copyright (c) 2001, Jonathan Revusky
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice, 
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation 
and/or other materials provided with the distribution.

The name of Jonathan Revusky may not be used to endorse or promote 
products derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE. 
*/

package com.revusky.oreo.predicates;

import com.revusky.oreo.*;
import com.revusky.oreo.util.Assert;
import java.util.*;

/**
 * A Predicate that represents a simple comparison filter.
 */

public class SimplePredicate extends Predicate {

    public static final int IS_NULL = 1;
    public static final int EQUALS = 2;
    public static final int GREATER_THAN = 3;
    public static final int LESS_THAN=4;
    public static final int LIKE = 5;
    public static final int NOT_NULL = -1;
    public static final int NOT_EQUALS = -2;
    public static final int LESS_THAN_EQUALS = -3;
    public static final int GREATER_THAN_EQUALS = -4;
    public static final int NOT_LIKE = -5;


    final String recType, fieldName;
    final Object comparisonValue;
    final Comparator comparator;
    final int comparisonType;
    boolean caseInsensitive;

    private String sqlQuery;

    protected SimplePredicate(int comparisonType, 
                           String recType,
                           String fieldName, 
                           Object comparisonValue, 
                           Comparator comparator,
                           boolean caseInsensitive) {
        if (comparisonType == EQUALS && comparisonValue == null)
            comparisonType = IS_NULL;
        if (comparisonType == NOT_EQUALS && comparisonValue == null)
            comparisonType = NOT_NULL;
        this.comparisonType = comparisonType;
        this.comparisonValue = caseInsensitive ? ((String) comparisonValue).toLowerCase() :  comparisonValue;
        this.fieldName = fieldName;
        this.comparator = comparator;
        this.recType = recType;
        this.caseInsensitive = caseInsensitive;
        deduceSQLQuery();
    }

    protected SimplePredicate(int comparisonType, 
                           String recType,
                           String fieldName, 
                           Object comparisonValue, 
                           Comparator comparator) {
        this(comparisonType, recType, fieldName, comparisonValue, comparator, false);
    }


    private void deduceSQLQuery() {
        String literal = SQLQueryUtil.escapeString(comparisonValue);
        String condition = null;
        String tableName = null;
        if (recType != null) {
            try {
                RecordDescriptor metadata = DataRegistryImpl.getDefaultRegistry().getMetadata(recType);
                tableName = metadata.getTableName();
            } catch (Exception e) {
                Assert.rethrow(e);
            }
        }
        switch (comparisonType) {
            case IS_NULL :
                condition = " IS ";
                break;
            case NOT_NULL : 
                condition = " IS NOT ";
                break;
            case EQUALS :
                condition = " = ";
                break;
            case NOT_EQUALS :
                condition = " != ";
                break;
            case  GREATER_THAN : 
                condition = " > ";
                break;
            case  GREATER_THAN_EQUALS : 
                condition = " >= ";
                break;
            case  LESS_THAN : 
                condition = " < ";
                break;
            case  LESS_THAN_EQUALS : 
                condition = " <= ";
                break;
            case LIKE :
                condition = " LIKE ";
                break;
            case NOT_LIKE :
                condition = " NOT LIKE ";
                break;
        }
        String f = caseInsensitive ? " WHERE LOWER(" + fieldName + ")" : fieldName;
        sqlQuery = "SELECT * FROM " + tableName + " WHERE " + f + condition + literal;
    }


    public String sqlEquivalent() {
        return sqlQuery;
    }

    public boolean accept(Record rec) {
        if (recType != null && !rec.getType().equals(recType))
            return false;
        Object val = rec.get(fieldName);
        if (caseInsensitive && val !=null)
            val = ((String) val).toLowerCase();
        switch (comparisonType) {
            case IS_NULL : 
                return val == null;
            case NOT_NULL :
                return val != null;
            case EQUALS :
                return comparisonValue.equals(val);
            case NOT_EQUALS :
                return !comparisonValue.equals(val);
            case LESS_THAN :
                return compare(val, comparisonValue) < 0;
            case GREATER_THAN :
                return compare(val, comparisonValue) > 0;
            case LESS_THAN_EQUALS :
                return compare(val, comparisonValue) <=0;
            case GREATER_THAN_EQUALS :
                return compare(val, comparisonValue) >=0;
            case LIKE :
                return SQLQueryUtil.matchLikeString((String) val, (String) comparisonValue);
            case NOT_LIKE :
                return !SQLQueryUtil.matchLikeString((String) val, (String) comparisonValue);
        }
        throw new UnsupportedOperationException();
    }

    public Predicate getComplement() {
        if (complement == null) {
            complement = new SimplePredicate(-1 * comparisonType, 
                                             recType,
                                             fieldName,
                                             comparisonValue, 
                                             comparator);
            complement.complement = this;
        }
        return complement;
    }

    public String recType() {
        return recType;
    }

    private int compare(Object o1, Object o2) {
        if (comparator != null) {
            return comparator.compare(o1, o2);
        }
        return ((Comparable) o1).compareTo(o2);
    }
}
