/*
OREO Object-relational layer
Copyright (c) 2001, Jonathan Revusky
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

The name of Jonathan Revusky may not be used to endorse or promote
products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.oreo.xml;

import com.revusky.oreo.*;
import com.revusky.oreo.util.*;
import org.xml.sax.*;
import org.xml.sax.helpers.DefaultHandler;
import org.apache.xerces.parsers.SAXParser;
import java.io.*;
import java.util.*;
import java.util.zip.*;
import java.text.*;

/**
 * A class to be used when parsing the XML files.
 */
class CustomContentHandler extends DefaultHandler {

    final DateFormat TIMESTAMP_FORMAT = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");

    MutableDataSource dataSource;
    Record currentRec, oldRec, newRec;
    FieldDescriptor currentField;
    Date currentTimeStamp;
    int transactionCount;

    boolean inSnapshot,
            inTransaction,
            inUpdate,
            inInsertion,
            inDeletion, inLog, inField;
    int transactionIndex;

    CustomContentHandler(MutableDataSource dataSource) {
        this.dataSource = dataSource;
    }

    public void startElement (String uri, String localName,
                  String qName, Attributes attributes)
    throws SAXException {
        if (localName.equals("SNAPSHOT"))
           inSnapshot = true;
        else if (localName.equals("TIMESTAMP")) {
            String ts = attributes.getValue("VALUE");
            try {
                currentTimeStamp = (Date) TIMESTAMP_FORMAT.parse(ts);
            }
            catch (ParseException pe) {
                throw new SAXException(pe);
            }
        }
        else if (localName.equals("DATA_LOG")) {
           inLog = true;
            transactionCount = 0;
        }
        else if (localName.equals("TRANSACTION")) {
           inTransaction = true;
           transactionIndex = 0;
           String type = attributes.getValue("TYPE");
           if (type.equals("UPDATE"))
              inUpdate = true;
           else if (type.equals("INSERTION"))
              inInsertion = true;
           else if (type.equals("DELETION"))
               inDeletion = true;
        }
        else if (localName.equals("RECORD")) {
            if (inTransaction)
               ++transactionIndex;
            String type = attributes.getValue("TYPE");
            Record rec = newRecord(type);
            currentRec = rec;
            if (transactionIndex == 1)
                oldRec = rec;
            else if (transactionIndex == 2)
                newRec = rec;
        }
        else if (localName.equals("FIELD")) {
            inField = true;
            String fieldName = attributes.getValue("NAME");
            currentField = currentRec.getMetadata().getField(fieldName);
            if (currentField == null)
               throw new SAXException("Unknown field: " + fieldName + " for record type: " + currentRec.getType());
        }
    }

    public void endElement (String uri, String localName, String qName)
    throws SAXException {
        if (localName.equals("SNAPSHOT"))
           inSnapshot = false;
        else if (localName.equals("DATA_LOG")) {
           inLog = false;
        }
        else if (localName.equals("TRANSACTION")) {
            transactionCount++;
            int eventType = -1;
            DataEvent event = null;
            if (inDeletion) {
                event = new DataEvent(dataSource,
                             DataEvent.RECORD_DELETED,
                             null,
                             currentRec,
                             currentTimeStamp);
            }
            else if (inInsertion) {
                event = new DataEvent(dataSource,
                             DataEvent.RECORD_INSERTED,
                             currentRec,
                             null,
                             currentTimeStamp);
            }
            else if (inUpdate) {
                event = new DataEvent(dataSource,
                             DataEvent.RECORD_UPDATED,
                             newRec,
                             oldRec,
                             currentTimeStamp);
            }
            if (event == null) {
                throw new SAXException("transaction type unknown");
            }
            try {
                dataSource.playEvent(event);
            } catch (IOException ioe) {
                throw new SAXException(ioe);
            }
            transactionIndex = 0;
            inTransaction = false;
            inDeletion = false;
            inInsertion = false;
            inUpdate = false;
        }
        else if (localName.equals("RECORD")) {
            if (!inTransaction) {
                try {
                    dataSource.playEvent(new DataEvent(dataSource,
                                                   DataEvent.RECORD_INSERTED,
                                                   currentRec,
                                                   null,
                                                   null));
                } catch (IOException ioe) {
                    throw new SAXException(ioe);
                }
            }
        }
        else if (localName.equals("FIELD")) {
            inField = false;
            currentField = null;
        }

    }

    public void characters (char ch[], int start, int length)
    throws SAXException {
        if (inField) {
            String s = new String(ch, start, length);
            if (currentField.getJavaType() == s.getClass()) {
                currentRec.set(currentField, s);
            }
            else try {
                Object val = currentField.getObjectFromString(currentRec, s);
                currentRec.set(currentField, val);
            } catch (IOException ioe) {
                throw new SAXException(ioe);
            }
        }
    }


    Record newRecord(String type) throws SAXException {
        try {
            Record record = DataRegistryImpl.getDefaultRegistry().getExemplar(type);
            if (record == null)
               throw new SAXException("RECORD TYPE: " + type + " not found");
            return record;
        }
        catch (IOException ioe) {
            throw new SAXException(ioe);
        }
    }
}

