/*
OREO Object-relational layer
Copyright (c) 2001, Jonathan Revusky
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

The name of Jonathan Revusky may not be used to endorse or promote
products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.oreo.xml;

import java.io.*;
import java.util.*;
import java.text.DateFormat;
import java.util.zip.*;
import java.text.SimpleDateFormat;
import org.xml.sax.*;
import org.apache.xerces.parsers.SAXParser;
import com.revusky.oreo.*;
import com.revusky.oreo.util.*;


/**
 * A DataListener implementation that logs data events
 * in an XML format
 */

public class DataLogger implements DataListener {

    private File skelFile, logFile, snapshotFile, snapshotFile_gz;
    private Date lastRotation;
    private boolean gzip= true; // always gzip snapshots and archived logs
    int logCount =0, compactFrequency = 1000;
    private MutableDataSource dataSource;
    private String encoding = "UTF-8";


    static final DateFormat FILENAME_APPEND_TIMESTAMP = new SimpleDateFormat("yyyy-MM-dd_HH-mm-ss");

    public DataLogger(MutableDataSource dataSource, Properties properties) throws IOException {
        this.dataSource = dataSource;
        String cfs = properties.getProperty("COMPACT_FREQUENCY");
        if (Util.hasContent(cfs)) {
            try {
                this.compactFrequency = Integer.parseInt(cfs);
            }
            catch (NumberFormatException nfe) {
                Logger.log("expecting number here, got: COMPACT_FREQUENCY=\"" + cfs + "\"");
            }
        }
        String path = properties.getProperty("DATA_PATH");
        String encodingString = properties.getProperty("ENCODING");
        if (Util.hasContent(encodingString)) {
            this.encoding = encodingString;
        }
        Assert.hasContent(path);
        File dataPath = new File(path);
        Assert.assert(dataPath.isAbsolute(), "DATA_PATH must be an absolute path");
        Assert.assert(dataPath.exists() || dataPath.mkdir(),
                      "Could not create directory " + path);
        Assert.assert(dataPath.isDirectory(), path + " is not a directory.");
        String filename = properties.getProperty("FILE_NAME");
        if (!Util.hasContent(filename))
            filename = dataSource.getName();
        if (filename.endsWith(".xml"))
            filename = filename.substring(0, filename.length() - 4);
        skelFile = new File(dataPath, filename + ".xml");
        logFile = new File(dataPath, filename + ".log");
        snapshotFile = new File(dataPath, filename + ".snapshot");
        snapshotFile_gz = new File(dataPath, filename + ".snapshot.gz");
        setup();
        dataSource.addDataListener(this);
    }

    public DataLogger(File skelFile, File snapshotFile, File logFile) throws IOException {
        this.skelFile = skelFile;
        this.logFile = logFile;
        this.snapshotFile = snapshotFile;
        String snf = snapshotFile.getName();
        if (snf.endsWith(".gz")) {
            snapshotFile_gz = snapshotFile;
            snapshotFile = new File(snapshotFile_gz.getParent(), snf.substring(0, snf.length() -3));
        }
        else
            snapshotFile_gz = new File(snapshotFile.getParent(), snf + ".gz");
        setup();
    }

    private void setup() throws IOException {
        if (!skelFile.exists())
            writeSkelFile();
        if (!logFile.exists())
            startNewLogFile();
        if (!snapshotFile_gz.exists() && !snapshotFile.exists()) {
            writeSnapShot();
        }
        File snap = gzip ? snapshotFile_gz : snapshotFile;
        Assert.assert(logFile.canWrite());
        Assert.assert(logFile.canRead());
        Assert.assert(snap.canWrite());
        Assert.assert(snap.canRead());
    }

    synchronized public void handleEvent(DataEvent event) throws IOException {
        if (timeToRotate(event)) {
            rotateLog();
        }
        FileOutputStream fos = new FileOutputStream(logFile.getAbsolutePath(), true);
        Writer writer = new OutputStreamWriter(fos, encoding);
//        FileWriter writer = new FileWriter(logFile.getAbsolutePath(), true);
        XMLOutput.outputXML(writer, event);
        writer.flush();
        writer.close();
        logCount++;
    }

    public void loadDataSourceFromFile() throws SAXException, IOException {
        SAXParser parser = new SAXParser();
        CustomContentHandler cch = new CustomContentHandler(dataSource);
        parser.setContentHandler(cch);
        parser.setEntityResolver(new GZEntityResolver(skelFile.getParentFile(), encoding));
        FileInputStream fis = new FileInputStream(skelFile);
        InputSource is = new InputSource(fis);
//        InputSource is = new InputSource(skelFile.getAbsolutePath());
        is.setPublicId(dataSource.getName());
        is.setSystemId(skelFile.getAbsolutePath());
        is.setEncoding(encoding);
        synchronized (dataSource) {
            parser.parse(is);
            fis.close();
        }
        logCount = cch.transactionCount;
    }

    synchronized public void rotateLog() throws IOException {
        if (gzip) {
            if (!snapshotFile_gz.exists() && snapshotFile.exists()) {
                Util.gzipFile(snapshotFile);
            }
        }
        else {
            if (!snapshotFile.exists() && snapshotFile_gz.exists()) {
                Util.gunzipFile(snapshotFile);
            }
        }
        File snfile = gzip ? snapshotFile_gz : snapshotFile;
        Assert.assert(snfile.exists(), "Missing snapshot file: " + snfile);
        Date lastModified = new Date(snfile.lastModified());
        String append = FILENAME_APPEND_TIMESTAMP.format(lastModified);
        String archivedName = snfile.getAbsolutePath();
        if (gzip) {
            archivedName = archivedName.substring(0, archivedName.length() - 3);
            archivedName += append;
            archivedName += ".gz";
        }
        else {
            archivedName += append;
        }
        snfile.renameTo(new File(archivedName));
        String archivedLogFile = logFile.getAbsolutePath() + append;
        File archLog = new File(archivedLogFile);
        logFile.renameTo(archLog);
        if (gzip)
            Util.gzipFile(archLog);
        writeSnapShot();
        startNewLogFile();
        logCount = 0;
    }

    private void startNewLogFile() throws IOException {
        Writer writer = new OutputStreamWriter(new FileOutputStream(logFile), encoding);
        XMLOutput.outputXMLTimeStamp(writer, new Date());
        writer.flush();
        writer.close();
    }

    private void writeSkelFile() throws IOException {
        Writer writer = new OutputStreamWriter(new FileOutputStream(skelFile), encoding);
        writer.write("<?xml version=\"1.0\"  encoding=\"" + encoding + "\" standalone=\"yes\" ?>\n");
        writer.write("<!DOCTYPE OREO_DATA [\n");
        writer.write("  <!ENTITY snapshot SYSTEM \"");
        writer.write(snapshotFile.getName());
        writer.write("\">\n");
        writer.write("  <!ENTITY logFile SYSTEM \"");
        writer.write(logFile.getName());
        writer.write("\">\n");
        writer.write("]>\n");
        writer.write("<OREO_DATA>\n");
        writer.write("&snapshot;");
        writer.write("\n");
        writer.write("<DATA_LOG>\n");
        writer.write("&logFile;");
        writer.write("\n");
        writer.write("</DATA_LOG>\n");
        writer.write("</OREO_DATA>");
        writer.flush();
        writer.close();
    }

    private void writeSnapShot() throws IOException {
        Collection recs = dataSource.select(null);
        Writer writer = gzip ?
                        (Writer) new OutputStreamWriter(new GZIPOutputStream(new FileOutputStream(snapshotFile_gz)), encoding) :
                        (Writer) new BufferedWriter(new OutputStreamWriter(new FileOutputStream(snapshotFile), encoding));
        XMLOutput.outputXML(writer, recs, "SNAPSHOT", true);
        writer.flush();
        writer.close();
    }

    protected boolean timeToRotate(DataEvent event) {
        return logCount >= compactFrequency;
    }
}
