/*
OREO Object-relational layer
Copyright (c) 2001, Jonathan Revusky
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

The name of Jonathan Revusky may not be used to endorse or promote
products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.oreo.xml;

import com.revusky.oreo.*;
import com.revusky.oreo.util.*;
import java.util.*;
import org.xml.sax.SAXException;
import java.io.IOException;

/**
 * A class that uses an XML-based persistence scheme.
 * All of the records are held in memory.
 * You must specify a DATA_PATH property when configuring
 * an instance of this data source. This must be a directory
 * on which your process has read/write privileges.
 * You may optionally specify the base filename. If none
 * is specified, it will use the name of the data source.
 * You can specify the optional parameter ENCODING
 * to indicate the encoding in which the XML files will be stored.
 * The default is UTF-8.
 * You may also specify a COMPACT_FREQUENCY parameter,
 * which says how frequently to rotate the transaction log.
 * The default (if no value is specified) is 1000.
 * When the log is rotated, this datasource archives
 * a snapshot+log in gzip format. Compression seems
 * logical, since the XML format is so extremely fat and redundant!
 */
public class XMLFileDataSource
extends AbstractMutableDataSource implements DataListener{

    private HashMap container = new HashMap();
    private Random rnd = new Random();

    public XMLFileDataSource() throws IOException {
    }

    public void init(Properties properties) throws IOException {
        super.init(properties);
        DataLogger logger = new DataLogger(this, properties);
        try {
            logger.loadDataSourceFromFile();
        }
        catch (SAXException se) {
            Exception e = se.getException();
            if (e instanceof IOException)
                throw (IOException) e;
            if (e instanceof RuntimeException)
                throw (RuntimeException) e;
            if (e == null)
                e = se;
            throw new AssertionFailedException(e);
        }
    }


    public synchronized void delete(Record record) throws IOException {
        markStale(record);
        container.remove(record.getPrimaryKey());
        DataEvent event = new DataEvent(this,
                                        DataEvent.RECORD_DELETED,
                                        null,
                                        record);
        fireDataEvent(event);
    }

    public synchronized void update(Record oldRec, Record newRec) throws IOException {
        Assert.assert(!newRec.isImmutable(),
                      "Expecting a mutable record.");
        Assert.assert(!newRec.isStale(),
                      "This record has been deleted. " +
                      "You cannot add it to the data source.");
        Assert.assert(oldRec.isImmutable(), "You can only update a record that is already initialized.");
        if (oldRec.isStale()) {
            throw new ModifiedRecordException(newRec, "Attempt to update a version that is not the latest one.");
        }
        Object primaryKey = newRec.getPrimaryKey();
        Assert.assert(oldRec.getPrimaryKey().equals(primaryKey));
        // The record is frozen, and hence, validated and immutable.
        freeze(newRec);
        markStale(oldRec);
        container.put(newRec.getPrimaryKey(), newRec);
        DataEvent event = new DataEvent(this,
                                        DataEvent.RECORD_UPDATED,
                                        newRec,
                                        oldRec);
        fireDataEvent(event);
    }

    public synchronized void insert(Record record) throws IOException {
        Object oldKey = record.getPrimaryKey();
        Object key = oldKey;
        if (key == null || "".equals(key)) {
            Class keyClass = record.getMetadata().getPrimaryKeyField().getJavaType();
            key = randomKey(keyClass);
            while (container.containsKey(key)) {
                key = randomKey(keyClass);
            }
        }
        else if (container.containsKey(key))
            throw new DuplicateRecordException(record,
                                               "The primary key "
                                               + key
                                               + " is already used.");
        record.setPrimaryKey(key);
        try {
            freeze(record);
        }
        catch (IOException ioe) {
            record.setPrimaryKey(oldKey);
            throw ioe;
        }
        container.put(key, record);
        DataEvent event = new DataEvent(
                                       this,
                                       DataEvent.RECORD_INSERTED,
                                       record,
                                       null);
        fireDataEvent(event);
    }

    private Object randomKey(Class cl) {
        if (cl == Integer.class) {
            return new Integer(Math.abs(rnd.nextInt()));
        }
        else if (cl == String.class) {
            int i = Math.abs(rnd.nextInt());
            return Integer.toString(i, 36);
        }
        else if (cl == Long.class) {
            return new Long(Math.abs(rnd.nextLong()));
        }
        else if (cl == Float.class) {
            return new Float(Math.abs(rnd.nextFloat()));
        }
        else if (cl == Double.class) {
            return new Double(Math.abs(rnd.nextDouble()));
        }
        throw new AssertionFailedException("Don't know how to generate a random object of class: " + cl);
    }

    public synchronized List select(RecordFilter filter) throws IOException {
        ArrayList result = new ArrayList();
        Iterator it = container.values().iterator();
        while (it.hasNext()) {
            Record rec = (Record) it.next();
            if (filter !=null && (!filter.accept(rec)))
                continue;
            result.add(rec);
        }
        result.trimToSize();
        return result;
    }

    public synchronized Record get(String type, Object key) throws IOException {
        if (key == null)
            return null;
        Record rec = (Record) container.get(key);
        if (rec == null && (key instanceof String)) {
            Object normkey = normalizeKey(type, (String) key);
            if (!key.equals(normkey))
                rec = (Record) container.get(normkey);
        }
        if (type != null
            && rec!= null
            && !type.equals(rec.getType())) {
            rec = null;
        }
        return rec;
    }

    public synchronized List keys(String type) throws IOException {
        Set allKeys = container.keySet();
        if (!Util.hasContent(type))
            return new ArrayList(allKeys);
        ArrayList result = new ArrayList();
        for (Iterator it = allKeys.iterator(); it.hasNext();) {
            Object key = it.next();
            Record rec = (Record) container.get(key);
            if (rec != null && rec.getType().equals(type))
                result.add(key);
        }
        result.trimToSize();
        return result;
    }

    public void handleEvent(DataEvent event) throws IOException {
        switch (event.type) {
            case DataEvent.RECORD_DELETED :
                delete(event.oldRecord);
                break;
            case DataEvent.RECORD_UPDATED :
                update(event.oldRecord, event.newRecord);
                break;
            case DataEvent.RECORD_INSERTED :
                insert(event.newRecord);
                break;
            default :
                throw new AssertionFailedException("Do not know how to handle this kind of event.");
        }
    }

    public void playEvent(DataEvent event) throws IOException {
        switch (event.type) {
            case DataEvent.RECORD_DELETED :
                container.remove(event.oldRecord.getPrimaryKey());
                break;
            case DataEvent.RECORD_UPDATED :
                container.remove(event.oldRecord.getPrimaryKey());
            case DataEvent.RECORD_INSERTED :
                Record rec = event.newRecord;
                freeze(rec);
                container.put(rec.getPrimaryKey(), rec);
                break;
            default :
                throw new AssertionFailedException("Do not know how to play this kind of event.");
        }
    }
}

