/*
OREO Object-relational layer
Copyright (c) 2001, Jonathan Revusky
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

The name of Jonathan Revusky may not be used to endorse or promote
products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

package com.revusky.oreo.xml;

import java.io.*;
import java.util.*;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import com.revusky.oreo.*;
import com.revusky.oreo.util.*;

public class XMLOutput {

    static public void outputXML(Writer writer, DataEvent event) throws IOException {
        String type = null;
        switch (event.type) {
            case DataEvent.RECORD_INSERTED :
                type = "INSERTION";
                break;
            case DataEvent.RECORD_DELETED :
                type = "DELETION";
                break;
            case DataEvent.RECORD_UPDATED :
                type = "UPDATE";
                break;
        }
        writer.write("\n<TRANSACTION TYPE=\"");
        writer.write(type);
        writer.write("\">\n");
        outputXMLTimeStamp(writer, event.getTimeStamp());
        if (event.oldRecord != null)
            outputXML(writer, event.oldRecord);
        if (event.type == DataEvent.RECORD_UPDATED)
           writer.write("<!-- replaced by -->");
        if (event.newRecord != null)
            outputXML(writer, event.newRecord);
        writer.write("</TRANSACTION>\n");
    }

    static public void outputXMLTimeStamp(Writer writer, Date d) throws IOException {
        writer.write("<TIMESTAMP VALUE=\"");
        writer.write(TIMESTAMP_FORMAT.format(d));
        writer.write("\" />");
    }



    /**
     * utility routine to output a set of records to
     * a character stream in XMl format.
     * @param elementName The name of the element to enclose the set of records in (may be null)
     */
    static public void outputXML(Writer writer, Collection recs, String elementName, boolean outputTimeStamp) throws IOException {
        if (Util.hasContent(elementName)) {
            writer.write("\n<");
            writer.write(elementName);
            writer.write('>');
        }
        if (outputTimeStamp) {
            outputXMLTimeStamp(writer, new Date());
        }
        for (Iterator it = recs.iterator(); it.hasNext();) {
            Record rec = (Record) it.next();
            outputXML(writer, rec);
        }
        if (Util.hasContent(elementName)) {
            writer.write("\n</");
            writer.write(elementName);
            writer.write(">\n");
        }
    }

    /**
     * utility routine to output a record as an XML element
     * to a character stream.
     */

    static public void outputXML(Writer writer, Record rec) throws IOException {
        writer.write("\n<RECORD TYPE=\"");
        writer.write(rec.getType());
        writer.write("\">");
        RecordDescriptor metadata  = rec.getMetadata();
        for (Enumeration fields = metadata.getFieldDescriptors(); fields.hasMoreElements();) {
            FieldDescriptor field = (FieldDescriptor) fields.nextElement();
            Object val = rec.get(field);
            if (val == null || "".equals(val))
                continue;
            writer.write("\n\t<FIELD NAME=\"");
            writer.write(field.getName());
            writer.write("\">");
            String rawString = (val instanceof String) ? (String) val : field.valueToString(val);
            String xmlString = convertToEntities(rawString);
            writer.write(xmlString);
            writer.write("</FIELD>");
        }
        writer.write("\n</RECORD>\n");
    }

    static public String convertToEntities(String s) {
        int slen = (s == null) ? 0 : s.length();
        if (slen == 0)
            return "";
        StringBuffer buf = new StringBuffer();
        StringTokenizer st = new StringTokenizer(s, escapedChars, true);
        while (st.hasMoreTokens()) {
            String tok = st.nextToken();
            int tlen = tok.length();
            if (tlen == slen)
                return s;
            int offset = (tlen == 1) ? escapedChars.indexOf(tok) : -1;
            if (offset >=0) {
                buf.append(entities[offset]);
            }
            else
                buf.append(tok);
        }
        return buf.toString();
    }

    /**
     * Utility routine to turn a Record into an
     * XML-formatted string.
     */

    static public String toXML(Record rec) {
        StringWriter sw = new StringWriter();
        try {
            outputXML(sw, rec);
        }
        catch (IOException e) {
            Assert.rethrow(e);
        }
        return sw.toString();
    }

    static final String escapedChars = "<>'\"&";
    static final String[] entities = new String[] {
        "&lt;", "&gt;", "&apos;", "&quot;", "&amp;"
    };

    static final DateFormat TIMESTAMP_FORMAT = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
}